from ._benchmark_bids import BenchmarkBid, BenchmarkBidApply, BenchmarkBidList
from ._benchmarkings import Benchmarking, BenchmarkingApply, BenchmarkingList
from ._bids import Bid, BidApply, BidList
from ._date_transformations import DateTransformation, DateTransformationApply, DateTransformationList
from ._day_ahead_bids import DayAheadBid, DayAheadBidApply, DayAheadBidList
from ._day_ahead_process import DayAheadProces, DayAheadProcesApply, DayAheadProcesList
from ._markets import Market, MarketApply, MarketList
from ._nord_pool_markets import NordPoolMarket, NordPoolMarketApply, NordPoolMarketList
from ._production_plan_time_series import (
    ProductionPlanTimeSeries,
    ProductionPlanTimeSeriesApply,
    ProductionPlanTimeSeriesList,
)
from ._rkom_bid_combinations import RKOMBidCombination, RKOMBidCombinationApply, RKOMBidCombinationList
from ._rkom_bids import RKOMBid, RKOMBidApply, RKOMBidList
from ._rkom_markets import RKOMMarket, RKOMMarketApply, RKOMMarketList
from ._rkom_process import RKOMProces, RKOMProcesApply, RKOMProcesList
from ._shop_transformations import ShopTransformation, ShopTransformationApply, ShopTransformationList

BenchmarkBidApply.model_rebuild()
BenchmarkingApply.model_rebuild()
BidApply.model_rebuild()
DayAheadBidApply.model_rebuild()
DayAheadProcesApply.model_rebuild()
RKOMBidApply.model_rebuild()
RKOMBidCombinationApply.model_rebuild()
RKOMProcesApply.model_rebuild()
ShopTransformationApply.model_rebuild()

__all__ = [
    "BenchmarkBid",
    "BenchmarkBidApply",
    "BenchmarkBidList",
    "Benchmarking",
    "BenchmarkingApply",
    "BenchmarkingList",
    "Bid",
    "BidApply",
    "BidList",
    "DateTransformation",
    "DateTransformationApply",
    "DateTransformationList",
    "DayAheadBid",
    "DayAheadBidApply",
    "DayAheadBidList",
    "DayAheadProces",
    "DayAheadProcesApply",
    "DayAheadProcesList",
    "Market",
    "MarketApply",
    "MarketList",
    "NordPoolMarket",
    "NordPoolMarketApply",
    "NordPoolMarketList",
    "ProductionPlanTimeSeries",
    "ProductionPlanTimeSeriesApply",
    "ProductionPlanTimeSeriesList",
    "RKOMBid",
    "RKOMBidApply",
    "RKOMBidList",
    "RKOMBidCombination",
    "RKOMBidCombinationApply",
    "RKOMBidCombinationList",
    "RKOMMarket",
    "RKOMMarketApply",
    "RKOMMarketList",
    "RKOMProces",
    "RKOMProcesApply",
    "RKOMProcesList",
    "ShopTransformation",
    "ShopTransformationApply",
    "ShopTransformationList",
]



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

__all__ = ["ProcessingLog", "ProcessingLogApply", "ProcessingLogList"]


class ProcessingLog(DomainModel):
    space: ClassVar[str] = "cogShop"
    error_message: Optional[str] = Field(None, alias="errorMessage")
    state: Optional[str] = None
    timestamp: Optional[str] = None


class ProcessingLogApply(DomainModelApply):
    space: ClassVar[str] = "cogShop"
    error_message: Optional[str] = None
    state: Optional[str] = None
    timestamp: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("cogShop", "ProcessingLog"),
            properties={
                "errorMessage": self.error_message,
                "state": self.state,
                "timestamp": self.timestamp,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        return InstancesApply(nodes, edges)


class ProcessingLogList(TypeList[ProcessingLog]):
    _NODE = ProcessingLog



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.market_configuration.data_classes import (
    RKOMBidCombination,
    RKOMBidCombinationApply,
    RKOMBidCombinationList,
)

from ._core import TypeAPI


class RKOMBidCombinationConfigurationsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBidCombination.bidConfigurations"},
        )
        if isinstance(external_id, str):
            is_rkom_bid_combination = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bid_combination)
            )

        else:
            is_rkom_bid_combinations = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_rkom_bid_combinations)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "RKOMBidCombination.bidConfigurations"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class RKOMBidCombinationsAPI(TypeAPI[RKOMBidCombination, RKOMBidCombinationApply, RKOMBidCombinationList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "RKOMBidCombination", "79f6abad4b6c04"),
            class_type=RKOMBidCombination,
            class_apply_type=RKOMBidCombinationApply,
            class_list=RKOMBidCombinationList,
        )
        self.bid_configurations = RKOMBidCombinationConfigurationsAPI(client)

    def apply(self, rkom_bid_combination: RKOMBidCombinationApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = rkom_bid_combination.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(RKOMBidCombinationApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(RKOMBidCombinationApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> RKOMBidCombination:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> RKOMBidCombinationList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> RKOMBidCombination | RKOMBidCombinationList:
        if isinstance(external_id, str):
            rkom_bid_combination = self._retrieve((self.sources.space, external_id))

            bid_configuration_edges = self.bid_configurations.retrieve(external_id)
            rkom_bid_combination.bid_configurations = [edge.end_node.external_id for edge in bid_configuration_edges]

            return rkom_bid_combination
        else:
            rkom_bid_combinations = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            bid_configuration_edges = self.bid_configurations.retrieve(external_id)
            self._set_bid_configurations(rkom_bid_combinations, bid_configuration_edges)

            return rkom_bid_combinations

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> RKOMBidCombinationList:
        rkom_bid_combinations = self._list(limit=limit)

        bid_configuration_edges = self.bid_configurations.list(limit=-1)
        self._set_bid_configurations(rkom_bid_combinations, bid_configuration_edges)

        return rkom_bid_combinations

    @staticmethod
    def _set_bid_configurations(
        rkom_bid_combinations: Sequence[RKOMBidCombination], bid_configuration_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in bid_configuration_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for rkom_bid_combination in rkom_bid_combinations:
            node_id = rkom_bid_combination.id_tuple()
            if node_id in edges_by_start_node:
                rkom_bid_combination.bid_configurations = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]



from __future__ import annotations

import logging
from pathlib import Path
from typing import Annotated

import cognite.client
import typer
from rich.console import Console
from rich.logging import RichHandler

from cognite import powerops
from cognite.powerops.clients import get_powerops_client

from . import resync
from ._models import MODEL_BY_NAME

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler(tracebacks_suppress=[cognite.client])]
)

log = logging.getLogger("rich")

app = typer.Typer()


def _version_callback(value: bool):
    if value:
        typer.echo(powerops.__version__)
        raise typer.Exit()


@app.callback()
def common(
    ctx: typer.Context,
    version: bool = typer.Option(None, "--version", callback=_version_callback),
):
    ...


@app.command(
    "plan",
    help="Preview the changes from the configuration files that `powerops apply` would make to the data model in CDF",
)
def plan(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
):
    log.info(f"Running plan on configuration files located in {path}")
    resync.plan(path, market)


@app.command("apply", help="Apply the changes from the configuration files to the data model in CDF")
def apply(
    path: Annotated[Path, typer.Argument(help="Path to configuration files")],
    market: Annotated[str, typer.Argument(help="Selected power market")],
):
    log.info(f"Running apply on configuration files located in {path}")

    resync.apply(path, market)


@app.command("deploy", help=f"Deploy the data model in CDF. Available models: {list(MODEL_BY_NAME.keys())}")
def deploy(
    models: Annotated[list[str], typer.Argument(help="The models to deploy")],
):
    client = get_powerops_client()
    for model_name in models:
        if model_name not in MODEL_BY_NAME:
            log.warning(f"Model {model_name} not found, skipping. Available models: {list(MODEL_BY_NAME.keys())}")
            continue
        log.info(f"Deploying {model_name} model...")

        model = MODEL_BY_NAME[model_name]
        result = client.cdf.data_modeling.graphql.apply_dml(
            # The removal of newlines is done to avoid a bug in the GraphQL API.
            model.id_,
            model.graphql.replace("\n", " "),
            model.name,
            model.description,
        )
        log.info(f"Deployed {model_name} model ({result.space}, {result.external_id}, {result.version})")


@app.command("show", help=f"Show the graphql schema of Power Ops model. Available models: {list(MODEL_BY_NAME.keys())}")
def show(
    model: Annotated[str, typer.Argument(help="The models to deploy")],
    remove_newlines: bool = typer.Option(
        False,
        "--remove-newlines",
        help="Remove newlines from the graphql schema. This is done when deploying the schema.",
    ),
):
    if model not in MODEL_BY_NAME:
        log.warning(f"Model {model} not found. Available models: {list(MODEL_BY_NAME.keys())}")
        typer.Exit()

    console = Console()
    graphql = MODEL_BY_NAME[model].graphql
    if remove_newlines:
        graphql = graphql.replace("\n", "")
        console.print(f"{graphql!r}")
    else:
        console.print(graphql)


def main():
    app()


if __name__ == "__main__":
    main()



from __future__ import annotations

from collections import Counter, defaultdict
from pathlib import Path
from typing import Any, Dict, Optional

from pydantic import BaseModel, ConfigDict, Field, field_validator
from pydantic_core.core_schema import ValidationInfo

from cognite.powerops.resync._settings import Settings
from cognite.powerops.resync.config.cogshop.shop_file_config import ShopFileConfig
from cognite.powerops.resync.config.market import BenchmarkingConfig, PriceScenario
from cognite.powerops.resync.config.market.dayahead import BidMatrixGeneratorConfig, BidProcessConfig
from cognite.powerops.resync.config.market.market import Market
from cognite.powerops.resync.config.market.rkom import RKOMBidCombinationConfig, RKOMBidProcessConfig, RkomMarketConfig
from cognite.powerops.resync.config.production.generator import GeneratorTimeSeriesMapping
from cognite.powerops.resync.config.production.plant import PlantTimeSeriesMapping
from cognite.powerops.resync.config.production.watercourse import WatercourseConfig
from cognite.powerops.resync.config.shared import TimeSeriesMapping
from cognite.powerops.resync.utils.serializer import load_yaml


class Config(BaseModel):
    @classmethod
    def load_yamls(cls, config_dir_path: Path) -> dict[str, Any]:
        """
        Loads all yaml files in the config directory and returns a dictionary with the file name as key and the
        content as value.

        Note the config_dir_path can be nested, all yaml files in the directory and subdirectories will be loaded.

        Parameters
        ----------
        config_dir_path: Path
            The path to the config directory

        Returns
        -------
            Dictionary with all yaml files loaded.
        """
        config: dict[str, Any] = {}
        field_names = set(cls.model_fields)
        for file_path in config_dir_path.glob("**/*.yaml"):
            if file_path.stem not in field_names:
                continue
            config[file_path.stem] = load_yaml(file_path)
        return config


class MarketConfig(Config):
    market: Market
    benchmarks: list[BenchmarkingConfig]
    price_scenario_by_id: dict[str, PriceScenario]

    bidprocess: list[BidProcessConfig]
    bidmatrix_generators: list[BidMatrixGeneratorConfig]

    rkom_bid_process: list[RKOMBidProcessConfig]
    rkom_bid_combination: Optional[list[RKOMBidCombinationConfig]] = None
    rkom_market: Optional[RkomMarketConfig] = None

    @field_validator("price_scenario_by_id")
    def no_duplicated_scenarios(cls, value: dict[str, PriceScenario]):
        scenarios_by_hash = defaultdict(list)
        for id_, s in value.items():
            scenarios_by_hash[hash(s.model_dump_json(exclude={"name"}))].append((id_, s))
        if duplicated := [s for s in scenarios_by_hash.values() if len(s) > 1]:
            sep = ") |\n\t\t\t ("
            raise ValueError(
                f"Scenarios  "
                f"({sep.join([', '.join([id_ for id_, _ in duplicate_set]) for duplicate_set in duplicated])}) "
                f"\nare duplicated."
            )
        return value

    @field_validator("rkom_bid_combination", mode="after")
    def valid_process_external_id(cls, value: list[RKOMBidCombinationConfig], values: ValidationInfo):
        valid_ids = {process_config.external_id for process_config in values.data["rkom_bid_process"]}
        for combination in value:
            for external_id_to_validate in combination.rkom_bid_config_external_ids:
                if external_id_to_validate not in valid_ids:
                    raise ValueError(
                        f"Reference to rkom bid process config in rkom_bid_combination yaml is wrong for "
                        f"{external_id_to_validate}. "
                        f"Possible references are: {[config.external_id for config in values['rkom_bid_process']]}"
                    )
        return value

    @field_validator("bidprocess", mode="after")
    def one_default_per_price_area(cls, value: list[BidProcessConfig]):
        default_configs = Counter(
            (bid_config.price_area_name for bid_config in value if bid_config.is_default_config_for_price_area)
        )

        if price_areas_with_multiple_default_configs := [
            price_area for price_area, count in default_configs.items() if count > 1
        ]:
            raise ValueError(
                f"Multiple default bid configs for price areas: {price_areas_with_multiple_default_configs}"
            )
        return value


class ProductionConfig(Config):
    dayahead_price_timeseries: Dict[str, str]
    watercourses: list[WatercourseConfig]
    generator_time_series_mappings: list[GeneratorTimeSeriesMapping] = None
    plant_time_series_mappings: list[PlantTimeSeriesMapping] = None

    @classmethod
    def load_yamls(cls, config_dir_path: Path) -> dict[str, Any]:
        config: dict[str, Any] = {}
        for field_name in cls.model_fields:
            if not (config_file_path := config_dir_path / f"{field_name}.yaml").exists():
                continue
            content = load_yaml(config_file_path)
            if field_name == "watercourses":
                for watercourse in content:
                    # Complete the paths for the yaml files
                    if all(key in watercourse for key in ["directory", "model_raw"]):
                        watercourse["yaml_raw_path"] = (
                            config_dir_path / watercourse["directory"] / watercourse["model_raw"]
                        )
                    if all(key in watercourse for key in ["directory", "model_processed"]):
                        watercourse["yaml_processed_path"] = (
                            config_dir_path / watercourse["directory"] / watercourse["model_processed"]
                        )
            config[field_name] = content
        return config


class CogShopConfig(Config):
    time_series_mappings: list[TimeSeriesMapping]
    watercourses_shop: list[ShopFileConfig]


class ReSyncConfig(BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    settings: Settings = Field(alias="constants")
    production: ProductionConfig
    market: MarketConfig
    cogshop: CogShopConfig

    @classmethod
    def from_yamls(cls, config_dir_path: Path, cdf_project: str) -> "ReSyncConfig":
        configs: dict[str, Any] = {}
        for field in cls.model_fields:
            if (config_file_path := config_dir_path / f"{field}.yaml").exists():
                configs[field] = load_yaml(config_file_path, encoding="utf-8")
            elif (config_subdir_path := config_dir_path / field).exists() and hasattr(
                cls.model_fields[field].annotation, "load_yamls"
            ):
                configs[field] = cls.model_fields[field].annotation.load_yamls(config_subdir_path)

        watercourse_directory_by_name = {}
        for watercourse in configs.get("production", {}).get("watercourses", []):
            if "directory" in watercourse and "name" in watercourse:
                watercourse_directory_by_name[watercourse["name"]] = watercourse["directory"]

        # Completing the paths for the shop files
        for shop_file in configs.get("cogshop", {}).get("watercourses_shop", []):
            if (watercourse_name := shop_file.get("watercourse_name")) and (
                directory := watercourse_directory_by_name.get(watercourse_name)
            ):
                if "file_path" in shop_file:
                    shop_file["file_path"] = config_dir_path / "production" / directory / shop_file["file_path"]
                # For backwards compatibility
                if "path" in shop_file:
                    shop_file["path"] = config_dir_path / "production" / directory / shop_file["path"]

        # Todo Hack to get cdf project into settings.
        if "settings" in configs:
            configs["settings"]["cdf_project"] = cdf_project
        elif "constants" in configs:
            # For backwards compatibility
            configs["constants"]["cdf_project"] = cdf_project

        return cls(**configs)



from ._generators import Generator, GeneratorApply, GeneratorList
from ._plants import Plant, PlantApply, PlantList
from ._price_areas import PriceArea, PriceAreaApply, PriceAreaList
from ._reservoirs import Reservoir, ReservoirApply, ReservoirList
from ._watercourses import Watercourse, WatercourseApply, WatercourseList

PlantApply.model_rebuild()
PriceAreaApply.model_rebuild()
WatercourseApply.model_rebuild()

__all__ = [
    "Generator",
    "GeneratorApply",
    "GeneratorList",
    "Plant",
    "PlantApply",
    "PlantList",
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "Reservoir",
    "ReservoirApply",
    "ReservoirList",
    "Watercourse",
    "WatercourseApply",
    "WatercourseList",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._date_transformations import DateTransformationApply

__all__ = ["ShopTransformation", "ShopTransformationApply", "ShopTransformationList"]


class ShopTransformation(DomainModel):
    space: ClassVar[str] = "power-ops"
    end: list[str] = []
    start: list[str] = []


class ShopTransformationApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    end: list[Union[str, "DateTransformationApply"]] = Field(default_factory=lambda: [], repr=False)
    start: list[Union[str, "DateTransformationApply"]] = Field(default_factory=lambda: [], repr=False)

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        for end in self.end:
            edge = self._create_end_edge(end)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(end, DomainModelApply):
                instances = end._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for start in self.start:
            edge = self._create_start_edge(start)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(start, DomainModelApply):
                instances = start._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return InstancesApply(nodes, edges)

    def _create_end_edge(self, end: Union[str, "DateTransformationApply"]) -> dm.EdgeApply:
        if isinstance(end, str):
            end_node_ext_id = end
        elif isinstance(end, DomainModelApply):
            end_node_ext_id = end.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(end)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ShopTransformation.end"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_start_edge(self, start: Union[str, "DateTransformationApply"]) -> dm.EdgeApply:
        if isinstance(start, str):
            end_node_ext_id = start
        elif isinstance(start, DomainModelApply):
            end_node_ext_id = start.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(start)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "ShopTransformation.start"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class ShopTransformationList(TypeList[ShopTransformation]):
    _NODE = ShopTransformation



from __future__ import annotations

from pydantic import BaseModel, Field

from cognite.powerops.resync.models._base import AssetType


class Connection(BaseModel):
    connection_type: str
    from_: str = Field(alias="from")
    from_type: str
    to: str
    to_type: str

    def to_from_any(self, from_: dict[str, AssetType], to: AssetType) -> AssetType | None:
        if self.to != to.name or self.to_type != to.type_:
            return None
        f = from_.get(self.from_)
        if f and self.from_type == f.type_:
            return f
        return None

    def from_to_any(self, from_: AssetType, to: dict[str, AssetType]) -> AssetType | None:
        if self.from_ != from_.name or self.from_type != from_.type_:
            return None
        f = to.get(self.to)

        if f and self.to_type == f.type_:
            return f
        return None



import datetime
from typing import Any, Optional, Union

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
WARNING = 30
WARN = WARNING
INFO = 20
DEBUG = 10
NOTSET = 0

_levelToName = {
    CRITICAL: "CRITICAL",
    ERROR: "ERROR",
    WARNING: "WARNING",
    INFO: "INFO",
    DEBUG: "DEBUG",
    NOTSET: "NOTSET",
}
_nameToLevel = {
    "CRITICAL": CRITICAL,
    "FATAL": FATAL,
    "ERROR": ERROR,
    "WARN": WARNING,
    "WARNING": WARNING,
    "INFO": INFO,
    "DEBUG": DEBUG,
    "NOTSET": NOTSET,
}


def _checkLevel(level: Any) -> int:
    if isinstance(level, int):
        rv = level
    elif str(level) == level:
        if level not in _nameToLevel:
            raise ValueError("Unknown level: %r" % level)
        rv = _nameToLevel[level]
    else:
        raise TypeError("Level not an integer or a valid string: %r" % (level,))
    return rv


class KnockoffLogger:
    def __init__(self, name: Optional[str] = None, level: Union[int, str] = NOTSET) -> None:
        self.name = name
        self.level = _checkLevel(level)

    def setLevel(self, level) -> None:
        self.level = _checkLevel(level)

    def isEnabledFor(self, level) -> bool:
        return level >= self.level

    def _log(self, level, msg, args, **kwargs) -> None:
        ts = datetime.datetime.utcnow().isoformat()
        level = _levelToName.get(_checkLevel(level), "")
        print(f"{ts} - {level} - {self.name}: {msg}")

    def log(self, level, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(level):
            self._log(level, msg, args, **kwargs)

    def debug(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(DEBUG):
            self._log(DEBUG, msg, args, **kwargs)

    def info(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(INFO):
            self._log(INFO, msg, args, **kwargs)

    def warning(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(WARNING):
            self._log(WARNING, msg, args, **kwargs)

    def error(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(ERROR):
            self._log(ERROR, msg, args, **kwargs)

    def critical(self, msg: str, *args, **kwargs) -> None:
        if self.isEnabledFor(CRITICAL):
            self._log(CRITICAL, msg, args, **kwargs)

    def exception(self, msg: str, *args, exc_info=True, **kwargs) -> None:
        self.error(msg, *args, exc_info=exc_info, **kwargs)


def getLogger(name: Optional[str] = None) -> KnockoffLogger:
    if not isinstance(name, str):
        raise TypeError("A logger name must be a string")
    return KnockoffLogger(name)



from datetime import datetime
from pathlib import Path
from typing import Callable
from uuid import uuid4

from cognite.client.data_classes import Event

from cognite.powerops.clients import get_powerops_client
from cognite.powerops.resync._logger import configure_debug_logging
from cognite.powerops.resync.config.resource_collection import ResourceCollection
from cognite.powerops.resync.config.resync_config import ReSyncConfig
from cognite.powerops.resync.to_models.transform import transform


def plan(path: Path, market: str, echo: Callable[[str], None] = None):
    echo = echo or print
    client = get_powerops_client()
    bootstrap_resources, config = _load_transform(market, path, client.cdf.config.project, echo)

    # 2.b - preview diff
    cdf_bootstrap_resources = bootstrap_resources.from_cdf(
        po_client=client, data_set_external_id=config.settings.data_set_external_id
    )

    echo(ResourceCollection.prettify_differences(bootstrap_resources.difference(cdf_bootstrap_resources)))


def apply(path: Path, market: str, echo: Callable[[str], None] = None):
    echo = echo or print
    client = get_powerops_client()
    collection, config = _load_transform(market, path, client.cdf.config.project, echo)

    # ! This should always stay at the bottom # TODO: consider wrapper
    collection.add(_create_bootstrap_finished_event(echo))

    # Step 3 - write bootstrap resources from diffs to CDF
    collection.write_to_cdf(
        client,
        config.settings.data_set_external_id,
        config.settings.overwrite_data,
    )


def _load_transform(market: str, path: Path, cdf_project: str, echo: Callable[[str], None]):
    # Step 1 - configure and validate config
    config = ReSyncConfig.from_yamls(path, cdf_project)
    configure_debug_logging(config.settings.debug_level)
    # Step 2 - transform from config to CDF resources and preview diffs
    echo(
        f"Running resync for data set {config.settings.data_set_external_id} "
        f"in CDF project {config.settings.cdf_project}"
    )
    bootstrap_resources = transform(config, market)
    return bootstrap_resources, config


def _create_bootstrap_finished_event(echo: Callable[[str], None]) -> Event:
    """Creating a POWEROPS_BOOTSTRAP_FINISHED Event in CDF to signal that bootstrap scripts have been ran"""
    current_time = int(datetime.utcnow().timestamp() * 1000)  # in milliseconds
    event = Event(
        start_time=current_time,
        end_time=current_time,
        external_id=f"POWEROPS_BOOTSTRAP_FINISHED_{str(uuid4())}",
        type="POWEROPS_BOOTSTRAP_FINISHED",
        subtype=None,
        source="PowerOps bootstrap",
        description="Manual run of bootstrap scripts finished",
    )
    echo(f"Created status event '{event.external_id}'")

    return event



from __future__ import annotations

from typing import ClassVar

from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models._base import AssetType, NonAssetType
from cognite.powerops.resync.models.cdf_resources import CDFSequence

from .base import Bid, Market, Process, ShopTransformation


class RKOMBid(Bid):
    auction: str
    block: str
    method: str
    minimum_price: str
    price_premium: str
    product: str
    watercourse: str
    price_scenarios: str
    reserve_scenarios: str


class RKOMPlants(NonAssetType):
    plants: str


class RKOMProcess(Process):
    type_: ClassVar[str] = "POWEROPS"
    label: ClassVar[AssetLabel] = AssetLabel.RKOM_BID_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations used in RKOM bid generation processes"
    shop: ShopTransformation
    bid: RKOMBid
    process_events: list[str] = Field(default_factory=list)
    timezone: str
    rkom: RKOMPlants
    incremental_mapping: list[CDFSequence] = Field(default_factory=list)


class RKOMCombinationBid(NonAssetType):
    auction: str
    combination_name: str
    rkom_bid_configs: list[str]


class RKOMBidCombination(AssetType):
    type_: ClassVar[str] = "rkom_bid_combination_configuration"
    label: ClassVar[str] = AssetLabel.RKOM_BID_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations for which bids should be combined into a total RKOM bid form"
    bid: RKOMCombinationBid


class RKOMMarket(Market):
    start_of_week: int



from __future__ import annotations

import json
from abc import ABC
from typing import ClassVar, Iterable, Optional
from typing import Type as TypingType
from typing import TypeVar

from cognite.client.data_classes import Asset, Label, Relationship, TimeSeries
from cognite.client.data_classes.data_modeling.instances import EdgeApply, NodeApply
from pydantic import BaseModel, ConfigDict

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.clients.cogshop.data_classes._core import DomainModelApply, InstancesApply
from cognite.powerops.resync.models.cdf_resources import CDFFile, CDFSequence


class Type(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        return self._fields_of_type(CDFSequence)

    def files(self) -> list[CDFFile]:
        return self._fields_of_type(CDFFile)

    def _fields_of_type(self, type_: TypingType[_T_Type]) -> list[_T_Type]:
        output: list[_T_Type] = []
        for field_name in self.model_fields:
            value = getattr(self, field_name)
            if not value:
                continue
            elif isinstance(value, list) and isinstance(value[0], type_):
                output.extend(value)
            elif isinstance(value, type_):
                output.append(value)
        return output


_T_Type = TypeVar("_T_Type")


class AssetType(Type, ABC):
    type_: ClassVar[Optional[str]] = None
    label: ClassVar[AssetLabel]
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)
    parent_description: ClassVar[Optional[str]] = None
    name: str
    description: Optional[str] = None
    _external_id: Optional[str] = None
    _parend_external_id: Optional[str] = None

    @property
    def external_id(self) -> str:
        if self._external_id:
            return self._external_id
        return f"{self.type_}_{self.name}"

    @external_id.setter
    def external_id(self, value: str) -> None:
        self._external_id = value

    @property
    def parent_external_id(self) -> str:
        if self._parend_external_id:
            return self._parend_external_id
        return f"{self.type_}s"

    @parent_external_id.setter
    def parent_external_id(self, value: str) -> None:
        self._parend_external_id = value

    @property
    def parent_name(self):
        return self.parent_external_id.replace("_", " ").title()

    def relationships(self) -> list[Relationship]:
        relationships = []
        for field_name, field in self.model_fields.items():
            value = getattr(self, field_name)
            if not value:
                continue
            if isinstance(value, list) and value and isinstance(value[0], AssetType):
                for target in value:
                    relationships.append(self._create_relationship(target.external_id, "ASSET", target.type_))
            elif isinstance(value, AssetType):
                target_type = value.type_
                if self.type_ == "plant" and value.type_ == "reservoir":
                    target_type = "inlet_reservoir"
                relationships.append(self._create_relationship(value.external_id, "ASSET", target_type))
            elif any(cdf_type in str(field.annotation) for cdf_type in [CDFSequence.__name__, TimeSeries.__name__]):
                if TimeSeries.__name__ in str(field.annotation):
                    target_type = "TIMESERIES"
                elif CDFSequence.__name__ in str(field.annotation):
                    target_type = "SEQUENCE"
                else:
                    raise ValueError(f"Unexpected type {field.annotation}")

                if isinstance(value, list):
                    for target in value:
                        relationships.append(self._create_relationship(target.external_id, target_type, field_name))
                else:
                    relationships.append(self._create_relationship(value.external_id, target_type, field_name))
        return relationships

    def as_asset(self):
        metadata = {}
        for field_name, field in self.model_fields.items():
            if (
                any(cdf_type in str(field.annotation) for cdf_type in [CDFSequence.__name__, TimeSeries.__name__])
                or field_name in {"name", "description", "label", "parent_external_id"}
                or field.exclude
            ):
                continue
            value = getattr(self, field_name)
            if (
                value is None
                or isinstance(value, AssetType)
                or (isinstance(value, list) and value and isinstance(value[0], AssetType))
            ):
                continue
            if isinstance(value, list) and not value:
                continue
            if isinstance(value, NonAssetType):
                value = value.model_dump(exclude_unset=True)
                for k, v in value.items():
                    if isinstance(v, (dict, list)):
                        v = json.dumps(v)
                    metadata[f"{field_name}:{k}"] = v
                continue

            if isinstance(value, (dict, list)):
                value = json.dumps(value)
            metadata[field_name] = value

        return Asset(
            external_id=self.external_id,
            name=self.name,
            parent_external_id=self.parent_external_id,
            labels=[Label(self.label.value)],
            metadata=metadata if metadata else None,
            description=self.description,
        )

    def _create_relationship(
        self,
        target_external_id: str,
        target_cdf_type: str,
        target_type: str,
    ) -> Relationship:
        source_type = "ASSET"

        # The market model uses the suffix CDF type for the relationship label, while the core model does not.
        try:
            # Core Model
            label = RelationshipLabel(f"relationship_to.{target_type}")
        except ValueError:
            # Market Model
            label = RelationshipLabel(f"relationship_to.{target_type}_{target_cdf_type.lower()}")
            # In addition, the Market Model uses capitalised CDF types for the relationship type,
            # while the core model uses all upper cases.
            source_type = source_type.title()
            target_cdf_type = target_cdf_type.title()

        return Relationship(
            external_id=f"{self.external_id}.{target_external_id}",
            source_external_id=self.external_id,
            source_type=source_type,
            target_external_id=target_external_id,
            target_type=target_cdf_type,
            labels=[Label(external_id=label.value)],
        )


class NonAssetType(BaseModel, ABC):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)


class Model(BaseModel, ABC):
    def sequences(self) -> list[CDFSequence]:
        return [sequence for item in self._types() for sequence in item.sequences()]

    def files(self) -> list[CDFFile]:
        return [file for item in self._types() for file in item.files()]

    def _types(self) -> Iterable[Type]:
        for f in self.model_fields:
            if isinstance(items := getattr(self, f), list) and items and isinstance(items[0], Type):
                yield from items


class AssetModel(Model, ABC):
    root_asset: ClassVar[Asset]

    def assets(self) -> list[Asset]:
        return [item.as_asset() for item in self._asset_types()]

    def relationships(self) -> list[Relationship]:
        return [edge for item in self._asset_types() for edge in item.relationships()]

    def parent_assets(self) -> list[Asset]:
        if not self.root_asset:
            # return []
            raise ValueError("Root asset not set")

        def _to_name(external_id: str) -> str:
            parts = external_id.replace("_", " ").split(" ")
            if parts[0].lower() == "rkom":
                parts[0] = parts[0].upper()
            else:
                parts[0] = parts[0].title()
            # The replacing is to have avoid doing changes, even though this is inconsistent with the other parent
            # assets.
            return " ".join(parts).replace("Bid process", "Bid").replace("bid process", "bid")

        parent_and_description_ids = {
            (item.parent_external_id, item.parent_description) for item in self._asset_types()
        }

        return [self.root_asset] + [
            Asset(
                external_id=parent_id,
                name=_to_name(parent_id),
                parent_external_id=self.root_asset.external_id,
                description=description,
            )
            for parent_id, description in parent_and_description_ids
        ]

    def _asset_types(self) -> Iterable[AssetType]:
        yield from (item for item in self._types() if isinstance(item, AssetType))


class DataModel(Model, ABC):
    def instances(self) -> InstancesApply:
        nodes: dict[str, NodeApply] = {}
        edges: dict[str, EdgeApply] = {}
        for domain_model in self._domain_models():
            instance_applies = domain_model.to_instances_apply()
            # Caching in case recursive relationships are used.
            for node in instance_applies.nodes:
                if node.external_id not in nodes:
                    nodes[node.external_id] = node
            for edge in instance_applies.edges:
                if edge.external_id not in edges:
                    edges[edge.external_id] = edge

        return InstancesApply(nodes=list(nodes.values()), edges=list(edges.values()))

    def _domain_models(self) -> Iterable[DomainModelApply]:
        for f in self.model_fields:
            if isinstance(items := getattr(self, f), list) and items and isinstance(items[0], DomainModelApply):
                yield from items



from __future__ import annotations

from typing import ClassVar

from cognite.powerops.resync.config.market._core import Configuration


class Market(Configuration):
    parent_external_id: ClassVar[str] = "market_configurations"
    external_id: str
    name: str
    max_price: float = None
    min_price: float = None
    time_unit: str = None
    timezone: str
    tick_size: float = None
    trade_lot: float = None
    price_steps: int = None
    price_unit: str = None


MARKET_CONFIG_NORDPOOL_DAYAHEAD = Market(
    external_id="market_configuration_nordpool_dayahead",
    name="Nord Pool Day-ahead",
    max_price=4000,
    min_price=-500,
    time_unit="1h",
    timezone="Europe/Oslo",
    tick_size=0.1,
    trade_lot=0.1,
    price_steps=200,
    price_unit="EUR/MWh",
)
MARKET_BY_PRICE_AREA = {"NO2": "Dayahead", "NO1": "1", "NO3": "1", "NO5": "1"}



import os
from dataclasses import asdict, dataclass
from datetime import datetime
from typing import Dict, List, Optional, Union

import dacite
import pandas as pd
from cognite.client import CogniteClient

from cognite.powerops.preprocessor import knockoff_logging as logging
from cognite.powerops.preprocessor.aggregation import resample_and_aggregate
from cognite.powerops.preprocessor.data_classes.time_series_mapping import TimeSeriesMapping
from cognite.powerops.preprocessor.utils import (
    merge_dicts,
    ms_to_datetime,
    retrieve_is_step,
    retrieve_time_series_datapoints,
    save_dict_as_yaml,
    shift_datetime_str,
)

logger = logging.getLogger(__name__)


# ! TODO: allow optional?
@dataclass
class CogShopCase:
    time: Optional[dict]
    model: dict
    connections: List[dict]
    commands: Optional[List[str]]

    @classmethod
    def from_dict(cls, d: Union[dict, List[dict]]) -> "CogShopCase":
        if isinstance(d, list):
            d = merge_dicts(*d)
        return dacite.from_dict(data_class=cls, data=d)

    def to_dict(self) -> dict:
        """Convert to dict. Ignoring top level keys with None value"""
        as_dict = asdict(self)
        remove = [key for key, value in as_dict.items() if value is None]
        for key in remove:
            as_dict.pop(key)
        return as_dict

    def save_to_file(self, out_dir: str = "", file_name: str = "model_processed.yaml") -> None:
        file_path = os.path.join(out_dir, file_name)
        save_dict_as_yaml(file_path=file_path, d=self.to_dict())

    @property
    def resolution_dict(self) -> Dict[str, int]:
        return self.time["timeresolution"] if self.time else {}

    @staticmethod
    def series_to_shop_datapoints(s: pd.Series) -> Dict[datetime, float]:
        """Converts a timestamp indexed series to a dict of datetime value pairs"""
        return {timestamp.to_pydatetime(): value for timestamp, value in s.to_dict().items()}

    def insert_datapoints(
        self,
        object_type: str,
        instance: Union[int, str],
        attribute: str,
        datapoints: Union[float, pd.Series],
    ) -> None:
        if isinstance(datapoints, pd.Series):
            datapoints = self.series_to_shop_datapoints(datapoints)  # type: ignore[assignment]
        try:
            self.model[object_type][instance][attribute] = datapoints
        except KeyError:
            logger.debug(f"Creating '{object_type}.{instance}.{attribute}'")
            if object_type not in self.model:
                self.model[object_type] = {}
            if instance not in self.model[object_type]:
                self.model[object_type][instance] = {}
            self.model[object_type][instance][attribute] = datapoints
        logger.debug(f"Datapoints updated on '{object_type}.{instance}.{attribute}'.")

    def replace_time_series(
        self,
        client: CogniteClient,
        start: int,
        end: int,
        mappings: List[TimeSeriesMapping],
        dynamic_minute_offset: int = 0,
    ) -> None:
        logger.debug(f"Replacing time series data between {ms_to_datetime(start)} and {ms_to_datetime(end)}.")
        ts_to_datapoints = retrieve_time_series_datapoints(client=client, mappings=mappings, start=start, end=end)
        ts_is_step = retrieve_is_step(client=client, mappings=mappings)
        for mapping in mappings:
            external_id = mapping.time_series_external_id
            datapoints = ts_to_datapoints.get(external_id, pd.Series(dtype=float))  # type: ignore[arg-type]

            for transformation in mapping.transformations:
                datapoints = transformation.apply(
                    datapoints=datapoints,
                    model=self.model,
                    object_type=mapping.object_type,
                    instance=mapping.instance,
                    start=start,
                    end=end,
                    shift_minutes=dynamic_minute_offset,
                    client=client,
                )

            # NOTE: this check has to happen here since `transform` might remove or create datapoints
            if datapoints.empty:
                logger.warning(f"No datapoints found for {mapping.shop_model_path} {external_id}")
                continue

            if mapping.retrieve in ["START", "END"]:
                logger.debug(f"Converting {mapping.shop_model_path} from datapoint to value only!")
                # NOTE: must covert to native python int/float since numpy.int/float cannot be dumped to yaml nicely
                datapoints = datapoints.iloc[0].item()

            # Aggregate and resample time series based on cog_shop_case.time.timeresolution
            # If we want to read time resolution from a time series, that could potentially be done
            #   through base/incremental mappping that populates cog_shop_case.timeresolution
            datapoints = resample_and_aggregate(  # type: ignore[assignment]
                datapoints=datapoints,
                method=mapping.aggregation,
                resolution_dict=self.resolution_dict,
                is_step=ts_is_step[external_id],
            )

            self.insert_datapoints(
                object_type=mapping.object_type,
                instance=mapping.instance,
                attribute=mapping.attribute,
                datapoints=datapoints,
            )
        logger.debug("Finished replacing time series data.")

    def set_time(self, starttime: str, endtime: str, timeresolution: Dict[str, int]) -> None:
        self.time = {
            "starttime": starttime,
            "endtime": endtime,
            "timeunit": "minute",
            "timeresolution": timeresolution,
        }
        logger.info(f"Using time specification {self.time}")

    def set_default_time(self, starttime: str, endtime: str) -> None:
        logger.info(f"Using default time specification: {self.time}")
        default_timeresolution = {
            starttime: 60,
            shift_datetime_str(dt=starttime, days=3): 240,
        }
        self.set_time(starttime, endtime, default_timeresolution)

    def _set_some_commands(self) -> None:
        logger.warning("Setting TEMPORARY commands")
        self.commands = [
            "set time_delay_unit minute",
            "set ramping /on",
            "set bypass_loss /on",
            "set mipgap 0.001000",
            "set timelimit 600.000",
            "set reserve_ramping_cost 1",
            "set fcr_n_equality /on",
            "set reserve_slack_cost 1",
            "set reserve_min_capacity 0.02",
            "set dyn_seg /on",
            "set dyn_juncloss /on",
            "start sim 1",
            "start sim 1",
            "start sim 1",
            "set code /inc",
            "start sim 1",
            "start sim 1",
            "start sim 1",
            "start sim 1",
        ]



from pydantic import Field

from cognite.powerops.clients.cogshop.data_classes import (
    FileRefApply,
    MappingApply,
    ModelTemplateApply,
    TransformationApply,
)

from ._base import DataModel, Type
from .cdf_resources import CDFFile, CDFSequence


class BaseMapping(Type):
    watercourse_name: str
    mapping: list[CDFSequence] = Field(default_factory=list)


class OutputDefinition(Type):
    watercourse_name: str
    mapping: list[CDFSequence] = Field(default_factory=list)


class ShopFile(Type):
    watercourse_name: str
    file: CDFFile


class CogShopModel(DataModel):
    base_mappings: list[BaseMapping] = Field(default_factory=list)
    output_definitions: list[OutputDefinition] = Field(default_factory=list)
    shop_files: list[ShopFile] = Field(default_factory=list)
    model_templates: list[ModelTemplateApply] = Field(default_factory=list)
    mappings: list[MappingApply] = Field(default_factory=list)
    transformations: list[TransformationApply] = Field(default_factory=list)
    file_refs: list[FileRefApply] = Field(default_factory=list)



from ._cases import Case, CaseApply, CaseList
from ._commands_configs import CommandsConfig, CommandsConfigApply, CommandsConfigList
from ._file_refs import FileRef, FileRefApply, FileRefList
from ._mappings import Mapping, MappingApply, MappingList
from ._model_templates import ModelTemplate, ModelTemplateApply, ModelTemplateList
from ._processing_logs import ProcessingLog, ProcessingLogApply, ProcessingLogList
from ._scenarios import Scenario, ScenarioApply, ScenarioList
from ._transformations import Transformation, TransformationApply, TransformationList

CaseApply.model_rebuild()
MappingApply.model_rebuild()
ModelTemplateApply.model_rebuild()
ScenarioApply.model_rebuild()

__all__ = [
    "Case",
    "CaseApply",
    "CaseList",
    "CommandsConfig",
    "CommandsConfigApply",
    "CommandsConfigList",
    "FileRef",
    "FileRefApply",
    "FileRefList",
    "Mapping",
    "MappingApply",
    "MappingList",
    "ModelTemplate",
    "ModelTemplateApply",
    "ModelTemplateList",
    "ProcessingLog",
    "ProcessingLogApply",
    "ProcessingLogList",
    "Scenario",
    "ScenarioApply",
    "ScenarioList",
    "Transformation",
    "TransformationApply",
    "TransformationList",
]



from __future__ import annotations

from typing import ClassVar, Optional

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

__all__ = ["NordPoolMarket", "NordPoolMarketApply", "NordPoolMarketList"]


class NordPoolMarket(DomainModel):
    space: ClassVar[str] = "power-ops"
    max_price: Optional[float] = Field(None, alias="maxPrice")
    min_price: Optional[float] = Field(None, alias="minPrice")
    name: Optional[str] = None
    price_steps: Optional[int] = Field(None, alias="priceSteps")
    price_unit: Optional[str] = Field(None, alias="priceUnit")
    tick_size: Optional[float] = Field(None, alias="tickSize")
    time_unit: Optional[str] = Field(None, alias="timeUnit")
    timezone: Optional[str] = None
    trade_lot: Optional[float] = Field(None, alias="tradeLot")


class NordPoolMarketApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    max_price: Optional[float] = None
    min_price: Optional[float] = None
    name: Optional[str] = None
    price_steps: Optional[int] = None
    price_unit: Optional[str] = None
    tick_size: Optional[float] = None
    time_unit: Optional[str] = None
    timezone: Optional[str] = None
    trade_lot: Optional[float] = None

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "NordPoolMarket"),
            properties={
                "maxPrice": self.max_price,
                "minPrice": self.min_price,
                "priceSteps": self.price_steps,
                "priceUnit": self.price_unit,
                "tickSize": self.tick_size,
                "timeUnit": self.time_unit,
                "tradeLot": self.trade_lot,
            },
        )
        sources.append(source)

        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "Market"),
            properties={
                "name": self.name,
                "timezone": self.timezone,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        return InstancesApply(nodes, edges)


class NordPoolMarketList(TypeList[NordPoolMarket]):
    _NODE = NordPoolMarket



from __future__ import annotations

from typing import ClassVar, Union

from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models._base import AssetType, NonAssetType


class Market(AssetType):
    type_: ClassVar[str] = "market_configuration"
    label: ClassVar[AssetLabel] = AssetLabel.MARKET
    parent_description: ClassVar[str] = "Configurations used for different markets"
    timezone: str


class DateTransformation(NonAssetType):
    transformation: str
    args: list[Union[int, float, str]] = Field(default_factory=list)
    kwargs: dict[str, Union[int, float, str]] = Field(default_factory=dict)


class ShopTransformation(NonAssetType):
    starttime: str
    endtime: str


class Bid(NonAssetType):
    date: str


class Process(AssetType):
    ...



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._rkom_process import RKOMProcesApply

__all__ = ["RKOMBidCombination", "RKOMBidCombinationApply", "RKOMBidCombinationList"]


class RKOMBidCombination(DomainModel):
    space: ClassVar[str] = "power-ops"
    auction: Optional[str] = None
    bid_configurations: list[str] = Field([], alias="bidConfigurations")
    name: Optional[str] = None


class RKOMBidCombinationApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    auction: Optional[str] = None
    bid_configurations: list[Union[str, "RKOMProcesApply"]] = Field(default_factory=lambda: [], repr=False)
    name: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "RKOMBidCombination"),
            properties={
                "auction": self.auction,
                "name": self.name,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        for bid_configuration in self.bid_configurations:
            edge = self._create_bid_configuration_edge(bid_configuration)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(bid_configuration, DomainModelApply):
                instances = bid_configuration._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return InstancesApply(nodes, edges)

    def _create_bid_configuration_edge(self, bid_configuration: Union[str, "RKOMProcesApply"]) -> dm.EdgeApply:
        if isinstance(bid_configuration, str):
            end_node_ext_id = bid_configuration
        elif isinstance(bid_configuration, DomainModelApply):
            end_node_ext_id = bid_configuration.external_id
        else:
            raise TypeError(f"Expected str or RKOMProcessApply, got {type(bid_configuration)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "RKOMBidCombination.bidConfigurations"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class RKOMBidCombinationList(TypeList[RKOMBidCombination]):
    _NODE = RKOMBidCombination



from __future__ import annotations

from cognite.powerops.cdf_labels import AssetLabel, RelationshipLabel
from cognite.powerops.resync.config.resource_collection import ResourceCollection
from cognite.powerops.resync.config.resync_config import ReSyncConfig

from .to_cogshop_model import to_cogshop_model
from .to_market_model import to_market_model
from .to_production_model import to_production_model


def transform(
    config: ReSyncConfig,
    market_name: str,
) -> ResourceCollection:
    production_model = to_production_model(config.production)
    market_model = to_market_model(config.market, production_model.price_areas, market_name)
    cogshop_model = to_cogshop_model(config.cogshop, production_model.watercourses, config.settings.shop_version)

    settings = config.settings
    market_model.set_root_asset(
        settings.shop_service_url,
        settings.organization_subdomain,
        settings.tenant_id,
        production_model.root_asset.external_id,
    )

    labels = AssetLabel.as_label_definitions() + RelationshipLabel.as_label_definitions()
    collection = ResourceCollection()
    collection.add(labels)

    for model in [production_model, market_model, cogshop_model]:
        collection.add(model.sequences())
        collection.add(model.files())
    for asset_model in [production_model, market_model]:
        collection.add(asset_model.parent_assets())
        collection.add(asset_model.assets())
        collection.add(asset_model.relationships())
    for data_model in [cogshop_model]:
        collection.add(data_model.instances())

    return collection



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._generators import GeneratorApply
    from ._reservoirs import ReservoirApply

__all__ = ["Plant", "PlantApply", "PlantList"]


class Plant(DomainModel):
    space: ClassVar[str] = "power-ops"
    display_name: Optional[str] = Field(None, alias="displayName")
    feeding_fee: Optional[str] = Field(None, alias="feedingFee")
    generators: list[str] = []
    head_loss_factor: Optional[float] = Field(None, alias="headLossFactor")
    inlet_level: Optional[str] = Field(None, alias="inletLevel")
    inlet_reservoirs: list[str] = Field([], alias="inletReservoirs")
    name: Optional[str] = None
    ordering: Optional[int] = None
    outlet_level: Optional[float] = Field(None, alias="outletLevel")
    outlet_level_time_series: Optional[str] = Field(None, alias="outletLevelTimeSeries")
    p_max: Optional[float] = Field(None, alias="pMax")
    p_max_time_series: Optional[str] = Field(None, alias="pMaxTimeSeries")
    p_min: Optional[float] = Field(None, alias="pMin")
    p_min_time_series: Optional[str] = Field(None, alias="pMinTimeSeries")
    penstock_head_loss_factors: Optional[dict] = Field(None, alias="penstockHeadLossFactors")
    water_value: Optional[str] = Field(None, alias="waterValue")


class PlantApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    display_name: Optional[str] = None
    feeding_fee: Optional[str] = None
    generators: list[Union[str, "GeneratorApply"]] = Field(default_factory=lambda: [], repr=False)
    head_loss_factor: Optional[float] = None
    inlet_level: Optional[str] = None
    inlet_reservoirs: list[Union[str, "ReservoirApply"]] = Field(default_factory=lambda: [], repr=False)
    name: Optional[str] = None
    ordering: Optional[int] = None
    outlet_level: Optional[float] = None
    outlet_level_time_series: Optional[str] = None
    p_max: Optional[float] = None
    p_max_time_series: Optional[str] = None
    p_min: Optional[float] = None
    p_min_time_series: Optional[str] = None
    penstock_head_loss_factors: Optional[dict] = None
    water_value: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "Plant"),
            properties={
                "displayName": self.display_name,
                "feedingFee": self.feeding_fee,
                "headLossFactor": self.head_loss_factor,
                "inletLevel": self.inlet_level,
                "name": self.name,
                "ordering": self.ordering,
                "outletLevel": self.outlet_level,
                "outletLevelTimeSeries": self.outlet_level_time_series,
                "pMax": self.p_max,
                "pMaxTimeSeries": self.p_max_time_series,
                "pMin": self.p_min,
                "pMinTimeSeries": self.p_min_time_series,
                "penstockHeadLossFactors": self.penstock_head_loss_factors,
                "waterValue": self.water_value,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        for generator in self.generators:
            edge = self._create_generator_edge(generator)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(generator, DomainModelApply):
                instances = generator._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        for inlet_reservoir in self.inlet_reservoirs:
            edge = self._create_inlet_reservoir_edge(inlet_reservoir)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(inlet_reservoir, DomainModelApply):
                instances = inlet_reservoir._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return InstancesApply(nodes, edges)

    def _create_generator_edge(self, generator: Union[str, "GeneratorApply"]) -> dm.EdgeApply:
        if isinstance(generator, str):
            end_node_ext_id = generator
        elif isinstance(generator, DomainModelApply):
            end_node_ext_id = generator.external_id
        else:
            raise TypeError(f"Expected str or GeneratorApply, got {type(generator)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "Plant.generators"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )

    def _create_inlet_reservoir_edge(self, inlet_reservoir: Union[str, "ReservoirApply"]) -> dm.EdgeApply:
        if isinstance(inlet_reservoir, str):
            end_node_ext_id = inlet_reservoir
        elif isinstance(inlet_reservoir, DomainModelApply):
            end_node_ext_id = inlet_reservoir.external_id
        else:
            raise TypeError(f"Expected str or ReservoirApply, got {type(inlet_reservoir)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "Plant.inletReservoirs"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class PlantList(TypeList[Plant]):
    _NODE = Plant



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._transformations import TransformationApply

__all__ = ["Mapping", "MappingApply", "MappingList"]


class Mapping(DomainModel):
    space: ClassVar[str] = "cogShop"
    aggregation: Optional[str] = None
    path: Optional[str] = None
    retrieve: Optional[str] = None
    timeseries_external_id: Optional[str] = Field(None, alias="timeseriesExternalId")
    transformations: list[str] = []


class MappingApply(DomainModelApply):
    space: ClassVar[str] = "cogShop"
    aggregation: Optional[str] = None
    path: str
    retrieve: Optional[str] = None
    timeseries_external_id: Optional[str] = None
    transformations: list[Union[str, "TransformationApply"]] = Field(default_factory=lambda: [], repr=False)

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("cogShop", "Mapping"),
            properties={
                "aggregation": self.aggregation,
                "path": self.path,
                "retrieve": self.retrieve,
                "timeseriesExternalId": self.timeseries_external_id,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        for transformation in self.transformations:
            edge = self._create_transformation_edge(transformation)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(transformation, DomainModelApply):
                instances = transformation._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        return InstancesApply(nodes, edges)

    def _create_transformation_edge(self, transformation: Union[str, "TransformationApply"]) -> dm.EdgeApply:
        if isinstance(transformation, str):
            end_node_ext_id = transformation
        elif isinstance(transformation, DomainModelApply):
            end_node_ext_id = transformation.external_id
        else:
            raise TypeError(f"Expected str or TransformationApply, got {type(transformation)}")

        return dm.EdgeApply(
            space="cogShop",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("cogShop", "Mapping.transformations"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("cogShop", end_node_ext_id),
        )


class MappingList(TypeList[Mapping]):
    _NODE = Mapping



from typing import Optional

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteReadTimeout
from pydantic import BaseModel
from pydantic.alias_generators import to_camel

from cognite.powerops.utils.retry import retry


class GraphQlModel(BaseModel, alias_generator=to_camel):  # type: ignore
    ...


class Transformation(GraphQlModel):
    method: str
    arguments: Optional[str] = None


class TransformationItems(GraphQlModel):
    items: list[Transformation]


class Commands(GraphQlModel):
    commands: list[str]


class FileRef(GraphQlModel):
    file_external_id: str
    type: Optional[str] = None


class FileRefItems(GraphQlModel):
    items: list[FileRef]


class Mapping(GraphQlModel):
    path: str
    timeseries_external_id: Optional[str] = None
    transformations: Optional[TransformationItems] = None
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None


class MappingItems(GraphQlModel):
    items: list[Mapping]


class ModelTemplate(GraphQlModel):
    watercourse: str
    model: FileRef
    base_mappings: MappingItems


class Scenario(GraphQlModel):
    name: str
    model_template: ModelTemplate
    commands: Commands
    mappings_override: Optional[MappingItems] = None
    extra_files: Optional[FileRefItems] = None


class Case(GraphQlModel):
    start_time: str
    end_time: str
    scenario: Scenario


def render_query(space: str, case_external_id: str) -> str:
    query_template = """
    query MyQuery {{
        getCaseById(instance: {{space: "{space}", externalId: "{external_id}"}}) {{
            items {{
                startTime
                endTime
                scenario {{
                    name
                    extraFiles (first: 500) {{
                        items {{
                            fileExternalId
                            type
                        }}
                    }}
                    mappingsOverride (first: 500) {{
                        items {{
                            path
                            timeseriesExternalId
                            retrieve
                            aggregation
                            transformations {{
                                items {{
                                    arguments
                                    method
                                }}
                            }}
                        }}
                    }}
                    modelTemplate {{
                        watercourse
                        model {{
                            fileExternalId
                        }}
                        baseMappings (first: 500) {{
                            items {{
                                path
                                timeseriesExternalId
                                retrieve
                                aggregation
                                transformations {{
                                    items {{
                                        method
                                        arguments
                                    }}
                                }}
                            }}
                        }}
                    }}
                    commands {{
                        commands
                    }}
                }}
            }}
        }}
    }}
    """
    return query_template.format(space=space, external_id=case_external_id)


def parse_response(response) -> list[Case]:
    data: dict = response.json()["data"]
    query_key = tuple(data.keys())[0]
    items = data[query_key]["items"]
    return [Case.parse_obj(item) for item in items]


@retry(CogniteReadTimeout, tries=5, delay=1, backoff=2, jitter=(0.1, 0.3))
def query_fdm(
    client: CogniteClient,
    space: str,
    case_external_id: str,
    model_extenal_id: Optional[str] = None,
    model_version: Optional[int] = None,
):
    project = client.iam.token.inspect().projects[0].url_name
    model = space if model_extenal_id is None else model_extenal_id
    version = str(model_version) if model_version is not None else "1"

    url = f"/api/v1/projects/{project}/userapis/spaces/{space}/datamodels/{model}/versions/{version}/graphql"
    return client.post(url, json={"query": render_query(space, case_external_id)})


def get_case(
    client: CogniteClient,
    space: str,
    case_external_id: str,
    model_extenal_id: Optional[str] = None,
    model_version: Optional[int] = None,
) -> Case:
    response = query_fdm(
        client=client,
        space=space,  # type: ignore[arg-type]
        case_external_id=case_external_id,  # type: ignore[arg-type]
        model_extenal_id=model_extenal_id,  # type: ignore[arg-type]
        model_version=model_version,  # type: ignore[arg-type]
    )

    return parse_response(response)[0]



from __future__ import annotations

from pathlib import Path
from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from cognite.powerops.resync.config.shared import TimeSeriesMapping
from cognite.powerops.resync.utils.serializer import load_yaml


class Watercourse(BaseModel):
    name: str
    shop_penalty_limit: int = 42000


class WatercourseConfig(Watercourse):
    """
    Represents the configuration for a Watercourse

    Attributes:
        version: The version of the watercourse configuration.

    """

    version: str
    market_to_price_area: Dict[str, str]
    directory: str
    model_raw: str
    # ------------------------------------------------------------------
    yaml_raw_path: Path
    yaml_processed_path: Path  # TODO: not used here
    yaml_mapping_path: str = ""
    model_processed: str  # TODO: not used here
    model_mapping: Optional[str] = None
    tco_paths: Optional[List[str]] = None  # TODO: not used here - HEV specific
    rrs_ids: Optional[List[str]] = None
    hardcoded_mapping: Optional[TimeSeriesMapping] = None  # TODO: not used here
    hist_flow_timeseries: Optional[Dict[str, str]] = None  # TODO: not used here
    # ------------------------------------------------------------------
    production_obligation_ts_ext_ids: List[str] = Field(default_factory=list)
    plant_display_names_and_order: Dict[str, tuple[str, int]] = Field(default_factory=dict)
    reservoir_display_names_and_order: Dict[str, tuple[str, int]] = Field(default_factory=dict)
    water_value_based_method_time_series_csv_filename: Optional[str] = None

    @classmethod
    def from_yaml(cls, yaml_path: Path) -> list["WatercourseConfig"]:
        return [cls(**watercourse_raw) for watercourse_raw in load_yaml(yaml_path)]



from __future__ import annotations

from typing import ClassVar

from cognite.client.data_classes import TimeSeries
from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models._base import NonAssetType

from .base import Bid, Process, ShopTransformation


class BenchmarkBid(Bid):
    market_config_external_id: str


class ProductionPlanTimeSeries(NonAssetType):
    name: str
    series: list[TimeSeries]


class BenchmarkProcess(Process):
    type_: ClassVar[str] = "benchmarking_configuration"
    parent_description: ClassVar[str] = "Configurations used in benchmarking processes"
    description: str = "Configuration for benchmarking of day-ahead bidding"
    label: ClassVar[AssetLabel] = AssetLabel.DAYAHEAD_BIDDING_BENCHMARKING_CONFIG
    shop: ShopTransformation
    bid: BenchmarkBid
    production_plan_time_series: list[ProductionPlanTimeSeries] = Field(default_factory=list)
    benchmarking_metrics: dict[str, str] = Field(default_factory=dict)
    run_events: list[str] = Field(default_factory=list)



from typing import ClassVar, Optional

from cognite.client.data_classes import Asset
from pydantic.dataclasses import Field

from cognite.powerops.resync.models._base import AssetModel

from .base import Bid, DateTransformation, Market, Process, ShopTransformation
from .benchmark import BenchmarkBid, BenchmarkProcess, ProductionPlanTimeSeries
from .dayahead import DayAheadBid, DayAheadProcess, NordPoolMarket
from .rkom import RKOMBid, RKOMBidCombination, RKOMCombinationBid, RKOMMarket, RKOMPlants, RKOMProcess


class MarketModel(AssetModel):
    root_asset: ClassVar[Asset] = None
    markets: list[Market] = Field(default_factory=list)
    processes: list[Process] = Field(default_factory=list)
    combinations: list[RKOMBidCombination] = Field(default_factory=list)

    @classmethod
    def set_root_asset(
        cls, shop_service_url: str, organization_subdomain: str, tenant_id: str, core_root_asset_external_id: str
    ) -> None:
        if shop_service_url == "https://shop-staging.az-inso-powerops.cognite.ai/submit-run":
            customer = "cognite"
        else:
            customer = organization_subdomain

        cls.root_asset = Asset(
            parent_external_id=core_root_asset_external_id,
            external_id="configurations",
            name="Configurations",
            description="Configurations used for PowerOps",
            metadata={
                "shop_service_url": shop_service_url,
                "organization_subdomain": organization_subdomain,
                "customer": customer,
                "tenant_id": tenant_id,
            },
        )


__all__ = [
    "RKOMBid",
    "RKOMBidCombination",
    "RKOMMarket",
    "RKOMProcess",
    "RKOMCombinationBid",
    "RKOMPlants",
    "DayAheadBid",
    "NordPoolMarket",
    "DayAheadProcess",
    "BenchmarkBid",
    "BenchmarkProcess",
    "ProductionPlanTimeSeries",
    "Market",
    "DateTransformation",
    "ShopTransformation",
    "Bid",
    "Process",
    "MarketModel",
]



from __future__ import annotations

import re
from typing import Optional

import pandas as pd
from cognite.client.data_classes import Sequence, TimeSeries

from cognite.powerops.resync.config.production.connections import Connection
from cognite.powerops.resync.config.resync_config import ProductionConfig
from cognite.powerops.resync.models import production
from cognite.powerops.resync.models.cdf_resources import CDFSequence
from cognite.powerops.resync.utils.serializer import load_yaml

p_min_fallback = 0
p_max_fallback = 1e20

head_loss_factor_fallback = 0.0


def to_production_model(config: ProductionConfig) -> production.ProductionModel:
    """
    Create Assets for:
        - price_area,
        - watercourse
        - plant
        - reservoir
        - generator

    Create Relationships for:
        - price_area -> plant
        - price_area -> watercourse
        - watercourse -> plant
        - watercourse -> production obligation time series
        - plant -> reservoir
        - plant -> generator
        - generator -> generator efficiency
        - generator -> turbine efficiency

    Create sequences for
        - Generator efficiency
        - Turbine efficiency

    Create the sequence data for the
        - Generator efficiency
        - Turbine efficiency

    Parameters
    ----------
    config : List[CoreConfigs]
        The configurations for the core model

    Returns
    -------
    CoreModel
        An instance of the Core Data Model
    """
    plant_time_series_mappings_by_name = {
        mapping.plant_name: mapping for mapping in (config.plant_time_series_mappings or [])
    }
    start_stop_cost_time_series_by_generator = {
        mapping.generator_name: mapping.start_stop_cost for mapping in (config.generator_time_series_mappings or [])
    }

    model = production.ProductionModel()
    for watercourse_config in config.watercourses:
        watercourse = production.Watercourse(
            name=watercourse_config.name,
            shop=production.WaterCourseShop(penalty_limit=str(watercourse_config.shop_penalty_limit)),
            config_version=watercourse_config.version,
            model_file=watercourse_config.yaml_raw_path,
            processed_model_file=watercourse_config.yaml_processed_path,
            plants=[],
            production_obligation_time_series=[
                TimeSeries(external_id=id_) for id_ in watercourse_config.production_obligation_ts_ext_ids
            ],
        )
        model.watercourses.append(watercourse)

        shop_case = load_yaml(watercourse_config.yaml_raw_path, clean_data=True)

        for reservoir_name in shop_case["model"]["reservoir"]:
            reservoir = production.Reservoir(
                name=reservoir_name,
                **dict(
                    zip(
                        ["display_name", "ordering"],
                        watercourse_config.reservoir_display_names_and_order.get(
                            reservoir_name, (re.sub(r"\([0-9]+\)", "", reservoir_name), "999")
                        ),
                    )
                ),
            )
            model.reservoirs.append(reservoir)

        for generator_name, generator_attributes in shop_case["model"]["generator"].items():
            start_stop_cost = start_stop_cost_time_series_by_generator.get(generator_name)
            generator = production.Generator(
                name=generator_name,
                penstock=str(generator_attributes.get("penstock", "1")),
                p_min=float(generator_attributes.get("p_min", 0.0)),
                startcost=float(_get_single_value(generator_attributes.get("startcost", 0.0))),
                start_stop_cost_time_series=TimeSeries(external_id=start_stop_cost) if start_stop_cost else None,
            )
            x_col_name = "generator_power"
            y_col_name = "generator_efficiency"
            sequence = Sequence(
                external_id=f"{generator.external_id}_generator_efficiency_curve",
                name=f"{generator.name} generator efficiency curve",
                columns=[
                    {"valueType": "DOUBLE", "externalId": x_col_name},
                    {"valueType": "DOUBLE", "externalId": y_col_name},
                ],
            )
            data = generator_attributes["gen_eff_curve"]
            efficiency_curve = CDFSequence(
                sequence=sequence,
                content=pd.DataFrame(
                    {
                        x_col_name: data["x"],
                        y_col_name: data["y"],
                    },
                    dtype=float,
                ),
            )
            generator.generator_efficiency_curve = efficiency_curve
            data = generator_attributes["turb_eff_curves"]
            ref_col_name = "head"
            x_col_name = "flow"
            y_col_name = "turbine_efficiency"
            sequence = Sequence(
                external_id=f"{generator.external_id}_turbine_efficiency_curve",
                name=f"{generator.name} turbine efficiency curve",
                columns=[
                    {"valueType": "DOUBLE", "externalId": ref_col_name},
                    {"valueType": "DOUBLE", "externalId": x_col_name},
                    {"valueType": "DOUBLE", "externalId": y_col_name},
                ],
            )
            df = pd.DataFrame(
                {
                    ref_col_name: [entry["ref"] for entry in data for _ in range(len(entry["x"]))],
                    x_col_name: [item for entry in data for item in entry["x"]],
                    y_col_name: [item for entry in data for item in entry["y"]],
                },
                dtype=float,
            )

            turbine_efficiency_curve = CDFSequence(
                sequence=sequence,
                content=df,
            )

            generator.turbine_efficiency_curve = turbine_efficiency_curve
            model.generators.append(generator)

        generators_by_name = {generator.name: generator for generator in model.generators}
        plants = []
        for plant_name, attributes in shop_case["model"]["plant"].items():
            display_name, order = watercourse_config.plant_display_names_and_order.get(
                plant_name, (re.sub(r"\([0-9]+\)", "", plant_name), "999")
            )
            mapping = plant_time_series_mappings_by_name.get(plant_name)
            if mapping:
                mappings = dict(
                    water_value_time_series=TimeSeries(external_id=mapping.water_value)
                    if mapping.water_value
                    else None,
                    inlet_level_time_series=TimeSeries(external_id=mapping.inlet_reservoir_level)
                    if mapping.inlet_reservoir_level
                    else None,
                    outlet_level_time_series=TimeSeries(external_id=mapping.outlet_reservoir_level)
                    if mapping.outlet_reservoir_level
                    else None,
                    feeding_fee_time_series=TimeSeries(external_id=mapping.feeding_fee)
                    if mapping.feeding_fee
                    else None,
                    p_min_time_series=TimeSeries(external_id=mapping.p_min) if mapping.p_min else None,
                    p_max_time_series=TimeSeries(external_id=mapping.p_max) if mapping.p_max else None,
                    head_direct_time_series=TimeSeries(external_id=mapping.head_direct)
                    if mapping.head_direct
                    else None,
                )
            else:
                mappings = {}

            plant = production.Plant(
                name=plant_name,
                display_name=display_name,
                ordering=str(order),
                outlet_level=float(attributes.get("outlet_line", 0)),
                p_min=float(attributes.get("p_min", p_min_fallback)),
                p_max=float(attributes.get("p_max", p_max_fallback)),
                head_loss_factor=float(
                    attributes.get("main_loss", [head_loss_factor_fallback])[0]
                ),  # For some reason, SHOP defines this as a list, but we only need the first (and only) value
                penstock_head_loss_factors={
                    str(penstock_number): float(loss_factor)
                    for penstock_number, loss_factor in enumerate(
                        attributes.get("penstock_loss", [head_loss_factor_fallback]), start=1
                    )
                },
                **mappings,
            )
            all_connections = shop_case["connections"]
            inlet_reservoir_name = _plant_to_inlet_reservoir_breadth_first_search(
                plant_name, all_connections, {r.name for r in model.reservoirs}
            )
            selected_reservoir = next((r for r in model.reservoirs if r.name == inlet_reservoir_name), None)
            plant.inlet_reservoir = selected_reservoir

            parsed_connections = [Connection(**connection) for connection in all_connections]

            # Add the generators to the plant
            plant.generators = [
                g for connection in parsed_connections if (g := connection.to_from_any(generators_by_name, plant))
            ] + [g for connection in parsed_connections if (g := connection.from_to_any(plant, generators_by_name))]
            plants.append(plant)

            prod_area = str(list(attributes["prod_area"].values())[0])
            price_area_name = watercourse_config.market_to_price_area[prod_area]
            price_area = production.PriceArea(name=price_area_name)
            if price_area_name not in {a.name for a in model.price_areas}:
                if price_area_name in config.dayahead_price_timeseries:
                    price_area.dayahead_price_time_series = TimeSeries(
                        external_id=config.dayahead_price_timeseries[price_area_name]
                    )
                model.price_areas.append(price_area)
            price_area = next(a for a in model.price_areas if a.name == price_area_name)
            watercourse.plants.append(plant)
            price_area.plants.append(plant)
            if watercourse.name not in {w.name for w in price_area.watercourses}:
                price_area.watercourses.append(watercourse)

        model.plants.extend(plants)

    return model


def _get_single_value(value_or_time_series: float | dict) -> float:
    """Get the single value from a time series, or a value
    returns the value if value_or_time_series is a value, otherwise the first value in the time series

    Parameters
    ----------
    value_or_time_series : float | dict
        Either a simple numeric value, or a dictionary (time series, {datetime string: value})
    """
    if isinstance(value_or_time_series, dict):
        return next(iter(value_or_time_series.values()))
    return value_or_time_series


def _plant_to_inlet_reservoir_breadth_first_search(
    plant_name: str,
    all_connections: list[dict],
    reservoirs: set[str],
) -> Optional[str]:
    """Search for a reservoir connected to a plant, starting from the plant and searching breadth first.

    Parameters
    ----------
    plant_name : str
        The plant we want to find a connection from
    all_connections : list[dict]
        All connections in the model.
    reservoirs : dict
        All reservoirs in the model. Keys are reservoir names.

    Returns
    -------
    Optional[str]
        The name of the reservoir connected to the plant, or None if no reservoir was found.
    """
    queue = []
    for connection in all_connections:
        if (
            connection["to"] == plant_name and connection.get("to_type", "plant") == "plant"
        ):  # if to_type is specified, it must be "plant"
            queue.append(connection)
            break
    visited = []
    while queue:
        connection = queue.pop(0)
        if connection not in visited:
            # Check if the given connection is from a reservoir
            # If we have "from_type" we can check directly if the object is a reservoir
            try:
                if connection["from_type"] == "reservoir":
                    return connection["from"]
            # If we don't have "from_type" we have to check if the name of the object is in the
            # list of reservoirs
            except KeyError:
                if connection["from"] in reservoirs:
                    return connection["from"]

            visited.append(connection)
            for candidate_connection in all_connections:
                # if the candidate connection is extension from the current connection, traverse it
                if candidate_connection["to"] == connection["from"]:
                    queue.append(candidate_connection)
    return None



from __future__ import annotations

import hashlib
import logging
from typing import Union, cast

from cognite.client.data_classes import Asset, Event, LabelDefinition, Relationship
from cognite.client.data_classes._base import CogniteResource, CogniteResourceList
from cognite.client.data_classes.data_modeling import EdgeApply, NodeApply
from cognite.client.exceptions import CogniteAPIError
from deepdiff import DeepDiff
from pydantic import BaseModel, Extra

from cognite.powerops.clients.cogshop.data_classes import (
    FileRef,
    FileRefApply,
    Mapping,
    MappingApply,
    ModelTemplate,
    ModelTemplateApply,
    Transformation,
    TransformationApply,
)
from cognite.powerops.clients.cogshop.data_classes._core import DomainModel, DomainModelApply, InstancesApply
from cognite.powerops.clients.powerops_client import PowerOpsClient
from cognite.powerops.resync.config.cogshop.shop_file_config import ShopFileConfig
from cognite.powerops.resync.config.shared import ExternalId
from cognite.powerops.resync.models.cdf_resources import CDFFile, CDFSequence
from cognite.powerops.utils.cdf.calls import upsert_cognite_resources

logger = logging.getLogger(__name__)

AddableResourceT = Union[
    "CogniteResource",
    list["CogniteResource"],
    list[ShopFileConfig],
    list[DomainModelApply],
    InstancesApply,
    CDFFile,
    list[CDFFile],
    CDFSequence,
    list[CDFSequence],
]


def dump_cdf_resource(resource) -> dict:
    """Legacy or DM resource."""
    try:
        dump_func = resource.dump
    except AttributeError:
        dump_func = resource.dict
    return dump_func()


class ResourceCollection(BaseModel):
    # CDF resources
    assets: dict[ExternalId, Asset] = {}
    relationships: dict[ExternalId, Relationship] = {}
    label_definitions: dict[ExternalId, LabelDefinition] = {}
    events: dict[ExternalId, Event] = {}
    files: dict[ExternalId, CDFFile] = {}
    sequences: dict[ExternalId, CDFSequence] = {}

    # dm resources:
    nodes: dict[ExternalId, NodeApply] = {}
    edges: dict[ExternalId, EdgeApply] = {}
    model_templates: dict[ExternalId, ModelTemplateApply] = {}
    mappings: dict[ExternalId, MappingApply] = {}
    file_refs: dict[ExternalId, FileRefApply] = {}
    transformations: dict[ExternalId, TransformationApply] = {}

    class Config:
        arbitrary_types_allowed = True
        extra = Extra.forbid

    @property
    def all_cdf_resources(self) -> list[CogniteResource]:
        """Not including DM."""
        return [
            *list(self.assets.values()),
            *list(self.relationships.values()),
            *list(self.sequences.values()),
            *list(self.label_definitions.values()),
            *list(self.events.values()),
        ]

    def add(self, resources_to_append: AddableResourceT):
        # sort the resource by type and append to the correct list
        if isinstance(resources_to_append, (list, CogniteResourceList)):
            for resource in resources_to_append:
                self._add_resource(resource)
        elif isinstance(resources_to_append, InstancesApply):
            for node in resources_to_append.nodes:
                self._add_resource(node)
            for edge in resources_to_append.edges:
                self._add_resource(edge)
        else:
            self._add_resource(resources_to_append)

    def _add_resource(self, resource: CogniteResource | ShopFileConfig):
        # cdf
        if isinstance(resource, Asset):
            self.assets[resource.external_id] = resource
        elif isinstance(resource, Relationship):
            self.relationships[resource.external_id] = resource
        elif isinstance(resource, LabelDefinition):
            self.label_definitions[resource.external_id] = resource
        elif isinstance(resource, Event):
            self.events[resource.external_id] = resource
        elif isinstance(resource, CDFFile):
            self.files[resource.external_id] = resource
        elif isinstance(resource, CDFSequence):
            self.sequences[resource.external_id] = resource
        # dm
        elif isinstance(resource, NodeApply):
            self.nodes[resource.external_id] = resource
        elif isinstance(resource, EdgeApply):
            self.edges[resource.external_id] = resource
        elif isinstance(resource, ModelTemplateApply):
            self.model_templates[resource.external_id] = resource
        elif isinstance(resource, MappingApply):
            self.mappings[resource.external_id] = resource
        elif isinstance(resource, TransformationApply):
            self.transformations[resource.external_id] = resource
        elif isinstance(resource, FileRefApply):
            self.file_refs[resource.external_id] = resource
        else:
            raise ValueError(f"Unknown resource type: {type(resource)}")

    def write_to_cdf(
        self,
        po_client: PowerOpsClient,
        data_set_external_id: str,
        overwrite: bool = False,
    ):
        # CDF auth and data set
        data_set_id = po_client.cdf.data_sets.retrieve(external_id=data_set_external_id).id
        for resource in self.all_cdf_resources:
            resource.data_set_id = data_set_id

        cdf_api_by_type = {
            "label_definitions": po_client.cdf.labels,
            "assets": po_client.cdf.assets,
            "relationships": po_client.cdf.relationships,
            "events": po_client.cdf.events,
        }
        for resource_type, api in cdf_api_by_type.items():
            resources = list(getattr(self, resource_type).values())
            logger.debug(f"Processing {len(resources)} {resource_type}...")
            upsert_cognite_resources(api, resource_type, resources)

        po_client.cdf.data_modeling.instances.apply(
            nodes=list(self.nodes.values()), edges=list(self.edges.values()), replace=True
        )

        logger.debug(f"Processing {len(self.files)} files...")
        for file in self.files.values():
            po_client.cdf.files.upload_bytes(
                content=file.content, data_set_id=data_set_id, overwrite=overwrite, **file.meta.dump(camel_case=False)
            )

        logger.debug(f"Processing {len(self.sequences)} sequences...")
        upsert_cognite_resources(
            po_client.cdf.sequences, "sequences", [cdf_sequence.sequence for cdf_sequence in self.sequences.values()]
        )
        for sequence_external_id, cdf_sequence in self.sequences.items():
            if cdf_sequence.content.empty:
                logger.warning(f"Empty sequence dataframe for {sequence_external_id}")
                continue
            try:
                po_client.cdf.sequences.data.insert_dataframe(
                    dataframe=cdf_sequence.content, external_id=sequence_external_id
                )
            except CogniteAPIError as e:
                logger.warning(f"Failed to insert sequence dataframe for {sequence_external_id}: {e}")

    def difference(self, cdf: "ResourceCollection") -> dict[str, str]:
        local = self

        resource_diff: dict[str, str] = {}

        # First, we handle the cdf resources (everything except sequence data and shop files)
        for cdf_resource in [
            "assets",
            "relationships",
            "label_definitions",
            "events",
            "model_templates",
            "file_refs",
            "mappings",
            "transformations",
            "sequences",
            "files",
        ]:
            local_resources = {
                resource.external_id: dump_cdf_resource(resource)
                for resource in getattr(local, cdf_resource).values()
                if not (isinstance(resource, Event) and resource.type == "POWEROPS_BOOTSTRAP_FINISHED")
            }
            clean_local_resources_for_diff(local_resources)
            cdf_resources = {
                resource.external_id: dump_cdf_resource(resource)
                for resource in getattr(cdf, cdf_resource).values()
                if not (isinstance(resource, Event) and resource.type == "POWEROPS_BOOTSTRAP_FINISHED")
            }
            clean_cdf_resources_for_diff(cdf_resources)

            resource_diff[cdf_resource] = DeepDiff(cdf_resources, local_resources, ignore_string_case=True).pretty()

        # Then we handle the sequence content
        local_sequence_content = {external_id: data.content.to_dict() for external_id, data in local.sequences.items()}
        cdf_sequence_content = {external_id: data.content.to_dict() for external_id, data in cdf.sequences.items()}
        sequence_content_diff = {
            "sequence_content": DeepDiff(
                cdf_sequence_content, local_sequence_content, ignore_nan_inequality=True
            ).pretty()
        }

        return resource_diff | sequence_content_diff

    @classmethod
    def prettify_differences(cls, diff_per_resource_type: dict[str, str]) -> str:
        output_message = ""
        for resource_type, differences in diff_per_resource_type.items():
            if differences:
                output_message += "--------------------------------------------\n"
                output_message += f"Difference for resource type {resource_type}: \n"
                output_message += "--------------------------------------------\n"
                output_message += differences
                output_message += "\n"
        return output_message.replace("added to dictionary.", "missing in CDF.").replace(
            "removed from dictionary.", "missing locally."
        )

    @classmethod
    def from_cdf(
        cls,
        po_client: PowerOpsClient,
        data_set_external_id: str,
    ) -> "ResourceCollection":
        """
        Function that creates a BootstrapResourceCollection from a CDF data set (typically the bootstrap data set)
        """

        data_set_id = po_client.cdf.data_sets.retrieve(external_id=data_set_external_id).id
        bootstrap_resource_collection = ResourceCollection()

        # start by downloading all the resources from CDF
        for resource_type in [
            "assets",
            "relationships",
            "labels",
            "events",
        ]:
            # get the api for the resource type
            api = getattr(po_client.cdf, resource_type)
            # get all the resources from CDF
            resources = api.list(data_set_ids=[data_set_id], limit=None)
            # add the resources to the bootstrap resource collection
            bootstrap_resource_collection.add(resources)

        for resource_type in [
            "file_refs",
            "transformations",
            "mappings",
            "model_templates",
        ]:
            # get the api for the resource type
            api = getattr(po_client.cog_shop, resource_type)
            # get all the resources from CDF
            actual_resources = api.list(limit=None)
            # convert actual resources to their "apply" form, for comparison
            apply_resources = to_dm_apply(actual_resources)
            # add the resources to the bootstrap resource collection
            bootstrap_resource_collection.add(apply_resources)

        file_meta = po_client.cdf.files.list(data_set_ids=[data_set_id], limit=None)
        shop_files = []
        for f in file_meta:
            if f.metadata.get("md5_hash") is None:
                file_content = po_client.cdf.files.download_bytes(external_id=f.external_id)
                md5_hash = hashlib.md5(file_content.replace(b"\r\n", b"\n")).hexdigest()
                f.metadata["md5_hash"] = md5_hash

            shop_files.append(CDFFile(meta=f))
        bootstrap_resource_collection.add(shop_files)

        # then download the sequence data
        all_sequences = po_client.cdf.sequences.list(data_set_ids=[data_set_id], limit=None)

        for sequence in all_sequences:
            sequence_data = po_client.cdf.sequences.data.retrieve_dataframe(
                external_id=sequence.external_id, limit=None, start=0, end=-1
            )
            bootstrap_resource_collection.add(CDFSequence(sequence=sequence, content=sequence_data))

        return bootstrap_resource_collection


def to_dm_apply(instances: list[DomainModel] | DomainModel) -> list[DomainModelApply]:
    items = cast(list[DomainModel], list(instances))
    apply_items = []

    type_map = {
        ModelTemplate: ModelTemplateApply,
        Mapping: MappingApply,
        Transformation: TransformationApply,
        FileRef: FileRefApply,
    }

    for item in items:
        if type(item) not in type_map:
            raise NotImplementedError(f"Dont know how to convert {item!r}.")
        apply_type = type_map[type(item)]

        apply_items.append(
            apply_type(**{field: value for field, value in item.dict().items() if field in apply_type.__fields__})
        )
    return apply_items


def clean_cdf_resources_for_diff(cdf_resources: dict[ExternalId, dict]) -> None:
    """Remove fields that are not relevant for diffing"""
    dynamic_fields = ["last_updated_time", "created_time", "parent_id", "root_id", "data_set_id", "id"]
    for resource in cdf_resources.values():
        for field in dynamic_fields:
            resource.pop(field, None)
        # remove the metadata if it is empty
        metadata = resource.get("metadata", {})
        if metadata == {}:
            resource.pop("metadata", None)
        else:
            for key, val in metadata.items():
                if isinstance(val, float):
                    # convert float to str
                    metadata[key] = str(val)

        # For sequences remove the column fields createdTime, lastUpdatedTime and metadata for each column
        if resource.get("columns"):
            for column in resource["columns"]:
                column.pop("createdTime", None)
                column.pop("lastUpdatedTime", None)
                column.pop("metadata", None)


def clean_local_resources_for_diff(local_resources: dict[ExternalId, dict]) -> None:
    for resource in local_resources.values():
        metadata = resource.get("metadata", {})
        for key, val in metadata.items():
            if isinstance(val, (float, int)):
                # convert float to str
                metadata[key] = str(val)



import getpass
import logging
import os
from pathlib import Path
from typing import Any, Literal, Optional, Type

import pydantic

try:
    import tomllib
except ModuleNotFoundError:
    import tomli as tomllib  # py < 3.11

from pydantic import BaseModel, Field, field_validator
from pydantic_settings import BaseSettings, PydanticBaseSettingsSource, SettingsConfigDict

__all__ = ["Settings", "CogniteSettings", "PoweropsRunSettings"]


logger = logging.getLogger(__name__)


class CogniteSettings(BaseModel):
    login_flow: Literal["client_credentials", "interactive"]
    project: str
    cdf_cluster: str
    tenant_id: str
    client_id: str
    client_secret: Optional[str] = None

    client_name: str = Field(default_factory=lambda: getpass.getuser())
    authority_host_uri: str = "https://login.microsoftonline.com"

    @property
    def base_url(self) -> str:
        return f"https://{self.cdf_cluster}.cognitedata.com/"

    @property
    def scopes(self) -> list[str]:
        return [f"{self.base_url}.default"]

    @property
    def authority_uri(self) -> str:
        return f"{self.authority_host_uri}/{self.tenant_id}"

    @property
    def token_url(self) -> str:
        return f"https://login.microsoftonline.com/{self.tenant_id}/oauth2/v2.0/token"


class PoweropsRunSettings(pydantic.BaseModel):
    read_dataset: Optional[str] = None
    write_dataset: Optional[str] = None
    cogshop_version: Optional[str] = None

    @field_validator("cogshop_version", mode="before")
    def number_to_str(cls, v):
        return str(v) if isinstance(v, (int, float)) else v


class Settings(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="SETTINGS__", env_nested_delimiter="__")
    cognite: CogniteSettings = {}
    powerops: PoweropsRunSettings = {}

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        """Add `file_settings` to sources (loads settings.toml and .secrets.toml)."""
        return init_settings, env_settings, _file_settings, file_secret_settings


def _file_settings() -> dict[str, Any]:
    settings_files = filter(None, os.environ.get("SETTINGS_FILES", "settings.toml;.secrets.toml").split(";"))
    collected_data = {}
    for file_path in settings_files:
        try:
            data = tomllib.loads(Path(file_path).read_text())
        except FileNotFoundError:
            pass
        else:
            logger.debug(f"Loaded settings from '{file_path}'.")
            # Merge dictionaries, assume one level of nesting.
            for key in collected_data.keys() | data.keys():
                if key in collected_data and key in data:
                    collected_data[key] = collected_data[key] | data[key]
                elif key in data:
                    collected_data[key] = data[key]
    return collected_data



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.market_configuration.data_classes import Benchmarking, BenchmarkingApply, BenchmarkingList

from ._core import TypeAPI


class BenchmarkingPlanTimeSeriesAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Benchmarking.productionPlanTimeSeries"},
        )
        if isinstance(external_id, str):
            is_benchmarking = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_benchmarking)
            )

        else:
            is_benchmarkings = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list(
                "edge", limit=-1, filter=f.And(is_edge_type, is_benchmarkings)
            )

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Benchmarking.productionPlanTimeSeries"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class BenchmarkingsAPI(TypeAPI[Benchmarking, BenchmarkingApply, BenchmarkingList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "Benchmarking", "52a1afc9b44b9f"),
            class_type=Benchmarking,
            class_apply_type=BenchmarkingApply,
            class_list=BenchmarkingList,
        )
        self.production_plan_time_series = BenchmarkingPlanTimeSeriesAPI(client)

    def apply(self, benchmarking: BenchmarkingApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = benchmarking.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(BenchmarkingApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(BenchmarkingApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Benchmarking:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> BenchmarkingList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Benchmarking | BenchmarkingList:
        if isinstance(external_id, str):
            benchmarking = self._retrieve((self.sources.space, external_id))

            production_plan_time_series_edges = self.production_plan_time_series.retrieve(external_id)
            benchmarking.production_plan_time_series = [
                edge.end_node.external_id for edge in production_plan_time_series_edges
            ]

            return benchmarking
        else:
            benchmarkings = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            production_plan_time_series_edges = self.production_plan_time_series.retrieve(external_id)
            self._set_production_plan_time_series(benchmarkings, production_plan_time_series_edges)

            return benchmarkings

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> BenchmarkingList:
        benchmarkings = self._list(limit=limit)

        production_plan_time_series_edges = self.production_plan_time_series.list(limit=-1)
        self._set_production_plan_time_series(benchmarkings, production_plan_time_series_edges)

        return benchmarkings

    @staticmethod
    def _set_production_plan_time_series(
        benchmarkings: Sequence[Benchmarking], production_plan_time_series_edges: Sequence[dm.Edge]
    ):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in production_plan_time_series_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for benchmarking in benchmarkings:
            node_id = benchmarking.id_tuple()
            if node_id in edges_by_start_node:
                benchmarking.production_plan_time_series = [
                    edge.end_node.external_id for edge in edges_by_start_node[node_id]
                ]



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, Optional

from cognite.client.data_classes import Asset, TimeSeries
from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models._base import AssetModel, AssetType, NonAssetType
from cognite.powerops.resync.models.cdf_resources import CDFSequence


class Generator(AssetType):
    type_: ClassVar[str] = "generator"
    label = AssetLabel.GENERATOR
    p_min: float
    penstock: str
    startcost: float
    start_stop_cost_time_series: Optional[TimeSeries] = None
    generator_efficiency_curve: Optional[CDFSequence] = None
    turbine_efficiency_curve: Optional[CDFSequence] = None


class Reservoir(AssetType):
    type_: ClassVar[str] = "reservoir"
    label = AssetLabel.RESERVOIR
    display_name: str
    ordering: str


class Plant(AssetType):
    type_: ClassVar[str] = "plant"
    label = AssetLabel.PLANT
    display_name: str
    ordering: str
    head_loss_factor: float
    outlet_level: float
    p_min: float
    p_max: float
    penstock_head_loss_factors: dict
    generators: list[Generator] = Field(default_factory=list)
    inlet_reservoir: Optional[Reservoir] = None
    p_min_time_series: Optional[TimeSeries] = None
    p_max_time_series: Optional[TimeSeries] = None
    water_value_time_series: Optional[TimeSeries] = None
    feeding_fee_time_series: Optional[TimeSeries] = None
    outlet_level_time_series: Optional[TimeSeries] = None
    inlet_level_time_series: Optional[TimeSeries] = None
    head_direct_time_series: Optional[TimeSeries] = None


class WaterCourseShop(NonAssetType):
    penalty_limit: str


class Watercourse(AssetType):
    type_: ClassVar[str] = "watercourse"
    label = AssetLabel.WATERCOURSE
    shop: WaterCourseShop
    config_version: str = Field(exclude=True)
    model_file: Path = Field(exclude=True)
    processed_model_file: Path = Field(exclude=True)
    plants: list[Plant]
    production_obligation_time_series: list[TimeSeries] = Field(default_factory=list)


class PriceArea(AssetType):
    type_: ClassVar[str] = "price_area"
    label = AssetLabel.PRICE_AREA
    dayahead_price_time_series: Optional[TimeSeries] = None
    plants: list[Plant] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)


class ProductionModel(AssetModel):
    root_asset: ClassVar[Asset] = Asset(external_id="power_ops", name="PowerOps")
    price_areas: list[PriceArea] = Field(default_factory=list)
    watercourses: list[Watercourse] = Field(default_factory=list)
    plants: list[Plant] = Field(default_factory=list)
    generators: list[Generator] = Field(default_factory=list)
    reservoirs: list[Reservoir] = Field(default_factory=list)



from __future__ import annotations

from typing import ClassVar

from cognite.client import data_modeling as dm

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

__all__ = ["CommandsConfig", "CommandsConfigApply", "CommandsConfigList"]


class CommandsConfig(DomainModel):
    space: ClassVar[str] = "cogShop"
    commands: list[str] = []


class CommandsConfigApply(DomainModelApply):
    space: ClassVar[str] = "cogShop"
    commands: list[str]

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("cogShop", "CommandsConfig"),
            properties={
                "commands": self.commands,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        return InstancesApply(nodes, edges)


class CommandsConfigList(TypeList[CommandsConfig]):
    _NODE = CommandsConfig



from __future__ import annotations

import getpass
from pathlib import Path

from cognite.client import ClientConfig, CogniteClient
from cognite.client.credentials import OAuthClientCredentials

from ._api.cases import CasesAPI
from ._api.commands_configs import CommandsConfigsAPI
from ._api.file_refs import FileRefsAPI
from ._api.mappings import MappingsAPI
from ._api.model_templates import ModelTemplatesAPI
from ._api.processing_logs import ProcessingLogsAPI
from ._api.scenarios import ScenariosAPI
from ._api.transformations import TransformationsAPI


class CogShopClient:
    """
    CogShopClient

    Generated with:
        pygen = 0.11.4
        cognite-sdk = 6.5.8
        pydantic = 2.0.2

    Data Model:
        space: cogShop
        externalId: CogShop
        version: 1
    """

    def __init__(self, config: ClientConfig | None = None):
        client = CogniteClient(config)
        self.cases = CasesAPI(client)
        self.commands_configs = CommandsConfigsAPI(client)
        self.file_refs = FileRefsAPI(client)
        self.mappings = MappingsAPI(client)
        self.model_templates = ModelTemplatesAPI(client)
        self.processing_logs = ProcessingLogsAPI(client)
        self.scenarios = ScenariosAPI(client)
        self.transformations = TransformationsAPI(client)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> CogShopClient:
        base_url = f"https://{cdf_cluster}.cognitedata.com/"
        credentials = OAuthClientCredentials(
            token_url=f"https://login.microsoftonline.com/{tenant_id}/oauth2/v2.0/token",
            client_id=client_id,
            client_secret=client_secret,
            scopes=[f"{base_url}.default"],
        )
        config = ClientConfig(
            project=project,
            credentials=credentials,
            client_name=getpass.getuser(),
            base_url=base_url,
        )

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str) -> CogShopClient:
        import toml

        return cls.azure_project(**toml.load(file_path)["cognite"])



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._date_transformations import DateTransformationApply
    from ._markets import MarketApply
    from ._shop_transformations import ShopTransformationApply

__all__ = ["DayAheadBid", "DayAheadBidApply", "DayAheadBidList"]


class DayAheadBid(DomainModel):
    space: ClassVar[str] = "power-ops"
    date: list[str] = []
    is_default_config_for_price_area: Optional[bool] = Field(None, alias="isDefaultConfigForPriceArea")
    main_scenario: Optional[str] = Field(None, alias="mainScenario")
    market: Optional[str] = None
    name: Optional[str] = None
    price_area: Optional[str] = Field(None, alias="priceArea")
    price_scenarios: list[dict] = Field([], alias="priceScenarios")
    shop: Optional[str] = None


class DayAheadBidApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    date: list[Union[str, "DateTransformationApply"]] = Field(default_factory=lambda: [], repr=False)
    is_default_config_for_price_area: Optional[bool] = None
    main_scenario: Optional[str] = None
    market: Optional[Union[str, "MarketApply"]] = Field(None, repr=False)
    name: Optional[str] = None
    price_area: Optional[str] = None
    price_scenarios: list[dict] = []
    shop: Optional[Union[str, "ShopTransformationApply"]] = Field(None, repr=False)

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "DayAheadBid"),
            properties={
                "isDefaultConfigForPriceArea": self.is_default_config_for_price_area,
                "mainScenario": self.main_scenario,
                "priceArea": self.price_area,
                "priceScenarios": self.price_scenarios,
                "shop": {
                    "space": "power-ops",
                    "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
                },
            },
        )
        sources.append(source)

        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "Bid"),
            properties={
                "market": {
                    "space": "power-ops",
                    "externalId": self.market if isinstance(self.market, str) else self.market.external_id,
                },
                "name": self.name,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        for date in self.date:
            edge = self._create_date_edge(date)
            if edge.external_id not in cache:
                edges.append(edge)
                cache.add(edge.external_id)

            if isinstance(date, DomainModelApply):
                instances = date._to_instances_apply(cache)
                nodes.extend(instances.nodes)
                edges.extend(instances.edges)

        if isinstance(self.market, DomainModelApply):
            instances = self.market._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return InstancesApply(nodes, edges)

    def _create_date_edge(self, date: Union[str, "DateTransformationApply"]) -> dm.EdgeApply:
        if isinstance(date, str):
            end_node_ext_id = date
        elif isinstance(date, DomainModelApply):
            end_node_ext_id = date.external_id
        else:
            raise TypeError(f"Expected str or DateTransformationApply, got {type(date)}")

        return dm.EdgeApply(
            space="power-ops",
            external_id=f"{self.external_id}:{end_node_ext_id}",
            type=dm.DirectRelationReference("power-ops", "Bid.date"),
            start_node=dm.DirectRelationReference(self.space, self.external_id),
            end_node=dm.DirectRelationReference("power-ops", end_node_ext_id),
        )


class DayAheadBidList(TypeList[DayAheadBid]):
    _NODE = DayAheadBid



from dataclasses import dataclass
from pathlib import Path

from cognite.client.data_classes.data_modeling import DataModelId

_POWEROPS_ROOT = Path(__file__).parent
GRAPHQL_SCHEMAS = _POWEROPS_ROOT / "data_models"


@dataclass
class PowerOpsModel:
    name: str
    description: str
    graphql_file: Path
    id_: DataModelId
    extra_types: list[Path] = None

    @property
    def graphql(self) -> str:
        graphql = self.graphql_file.read_text()
        if self.extra_types:
            graphql = "\n\n".join([extra.read_text() for extra in self.extra_types] + [graphql])

        return graphql


_SPACE = "power-ops"
_INPUT_TIMESERIES_MAPPING = GRAPHQL_SCHEMAS / "input_timeseries_mapping.graphql"
_VALUE_TRANSFORMATION = GRAPHQL_SCHEMAS / "value_transformation.graphql"

MODEL_BY_NAME: dict[str, PowerOpsModel] = {
    "production": PowerOpsModel(
        name="Production",
        description="The production model descripbes the physical assets such as watercourses, "
        "plants, and generators located in a price area.",
        graphql_file=GRAPHQL_SCHEMAS / "production.graphql",
        id_=DataModelId(_SPACE, "production", "1"),
    ),
    "market": PowerOpsModel(
        name="Market",
        description="The market model describes the different markets in an price area and their "
        "financial assets such as "
        "bids and processes. In addition, to benchmarking.",
        graphql_file=GRAPHQL_SCHEMAS / "market.graphql",
        id_=DataModelId(_SPACE, "market", "1"),
        extra_types=[_INPUT_TIMESERIES_MAPPING, _VALUE_TRANSFORMATION],
    ),
    "cogshop": PowerOpsModel(
        name="CogShop",
        description="The CogShop model describes the interaction between Cognite Data Fusion (CDF) and "
        "Sintef's SHOP algorithm. The scenario is used to determine which SHOP runs are"
        "executed daily and configuration of those SHOP runs.",
        graphql_file=GRAPHQL_SCHEMAS / "cogshop.graphql",
        id_=DataModelId(_SPACE, "cogshop", "1"),
        extra_types=[_INPUT_TIMESERIES_MAPPING, _VALUE_TRANSFORMATION],
    ),
}



from __future__ import annotations

import json
import logging
from datetime import datetime
from hashlib import md5
from pathlib import Path

import pandas as pd
import yaml
from cognite.client.data_classes import FileMetadata, Sequence

from cognite.powerops.clients.cogshop.data_classes import (
    FileRefApply,
    MappingApply,
    ModelTemplateApply,
    TransformationApply,
)
from cognite.powerops.resync.config.resync_config import CogShopConfig
from cognite.powerops.resync.models import cogshop
from cognite.powerops.resync.models.cdf_resources import CDFSequence
from cognite.powerops.resync.models.cogshop import CogShopModel
from cognite.powerops.resync.models.production import Watercourse
from cognite.powerops.resync.utils.serializer import load_yaml

logger = logging.getLogger(__name__)


def to_cogshop_model(config: CogShopConfig, watercourses: list[Watercourse], shop_version: str) -> CogShopModel:
    model = CogShopModel()

    for shop_config in config.watercourses_shop:
        file_content = Path(shop_config.file_path).read_bytes()
        shop_file = cogshop.ShopFile(
            watercourse_name=shop_config.watercourse_name,
            file=cogshop.CDFFile(
                meta=FileMetadata(
                    external_id=shop_config.external_id,
                    name=shop_config.external_id,
                    metadata={
                        "shop:type": shop_config.cogshop_file_type,
                        "shop:watercourse": shop_config.watercourse_name,
                        "shop:file_name": shop_config.file_path.stem,
                        "md5_hash": md5(file_content.replace(b"\r\n", b"\n")).hexdigest(),
                    },
                    source="PowerOps bootstrap",
                    mime_type="text/plain",
                ),
                content=file_content,
            ),
        )
        model.shop_files.append(shop_file)

    # TODO Fix the assumption that timeseries mappings and watercourses are in the same order
    for watercourse, mapping in zip(watercourses, config.time_series_mappings):
        #### Model File ####
        processed_model = _to_model_without_timeseries(
            yaml_raw_path=str(watercourse.model_file),
            # Todo Move this hardcoded configuration to a config file
            non_time_series_attributes_to_remove={"reservoir.start_vol", "generator.initial_state"},
            encoding="utf-8",
        )
        yaml_content = yaml.safe_dump(processed_model, allow_unicode=True, sort_keys=False)
        file_content = yaml_content.encode("utf-8")
        external_id = f"SHOP_{watercourse.name}_{watercourse.processed_model_file.stem}"
        model_file = cogshop.ShopFile(
            watercourse_name=watercourse.name,
            file=cogshop.CDFFile(
                meta=FileMetadata(
                    external_id=external_id,
                    name=external_id,
                    metadata={
                        "shop:type": "case",
                        "shop:watercourse": watercourse.name,
                        "shop:file_name": watercourse.processed_model_file.stem,
                        "md5_hash": md5(file_content.replace(b"\r\n", b"\n")).hexdigest(),
                    },
                    source="PowerOps bootstrap",
                    mime_type="text/plain",
                ),
                content=file_content,
            ),
        )
        model.shop_files.append(model_file)

        ##### Output definition #####
        external_id = f"SHOP_{watercourse.name.replace(' ', '_')}_output_definition"

        sequence = Sequence(
            name=external_id.replace("_", " "),
            description="Defining which SHOP results to output to CDF (as time series)",
            external_id=external_id,
            columns=[
                {"valueType": "STRING", "externalId": "shop_object_type"},
                {"valueType": "STRING", "externalId": "shop_attribute_name"},
                {"valueType": "STRING", "externalId": "cdf_attribute_name"},
                {"valueType": "STRING", "externalId": "unit"},
                {"valueType": "STRING", "externalId": "is_step"},
            ],
            metadata={
                "shop:watercourse": watercourse.name,
                "shop:type": "output_definition",
            },
        )
        # Only default mapping is used
        df = pd.DataFrame(
            [
                ("market", "sale_price", "price", "EUR/MWh", "True"),
                ("market", "sale", "sales", "MWh", "True"),
                ("plant", "production", "production", "MW", "True"),
                ("plant", "consumption", "consumption", "MW", "True"),
                ("reservoir", "water_value_global_result", "water_value", "EUR/Mm3", "True"),
                ("reservoir", "energy_conversion_factor", "energy_conversion_factor", "MWh/Mm3", "True"),
            ],
            columns=[c["externalId"] for c in sequence.columns],
        )

        output_definition = cogshop.OutputDefinition(
            watercourse_name=watercourse.name,
            mapping=[
                CDFSequence(
                    sequence=sequence,
                    content=df,
                )
            ],
        )
        model.output_definitions.append(output_definition)

        ##### Base Mapping #####
        external_id = f"SHOP_{watercourse.name}_base_mapping"
        sequence = Sequence(
            name=external_id.replace("_", " "),
            external_id=external_id,
            description="Mapping between SHOP paths and CDF TimeSeries",
            columns=mapping.column_definitions,
            metadata={
                "shop:watercourse": watercourse.name,
                "shop:type": "base_mapping",
            },
        )
        output_definition = cogshop.BaseMapping(
            watercourse_name=watercourse.name,
            mapping=[
                CDFSequence(
                    sequence=sequence,
                    content=mapping.to_dataframe(),
                )
            ],
        )
        model.base_mappings.append(output_definition)

        ############## Adding the Instances. ##############
        file_ref = FileRefApply(
            external_id=f"ModelTemplate_{watercourse.name}__FileRef_model",
            type="model",
            file_external_id=model_file.file.external_id,
        )
        model.file_refs.append(file_ref)
        transformations = []
        base_mappings = []
        for row in reversed(list(mapping)):
            row_ext_id = f"BM__{watercourse.name}__{row.shop_model_path}"

            # We can get duplicate mappings (same path). Only keep the last one (look is reversed):
            visited_ext_ids = set()
            if row_ext_id in visited_ext_ids:
                logger.warning(f"Duplicate base mapping: {row_ext_id}")
                continue
            visited_ext_ids.add(row_ext_id)

            row_transformations = []
            for transformation in reversed(row.transformations or []):
                row_transformations.append(
                    TransformationApply(
                        external_id=_make_ext_id(
                            watercourse.name,
                            row.shop_model_path,
                            transformation.transformation.name,
                            json.dumps(transformation.kwargs or {}),
                        ),
                        method=transformation.transformation.name,
                        arguments=json.dumps(transformation.kwargs or {}),
                    )
                )
            transformations.extend(row_transformations)

            mapping_apply = MappingApply(
                external_id=row_ext_id,
                path=row.shop_model_path,
                timeseries_external_id=row.time_series_external_id,
                transformations=[tr.external_id for tr in row_transformations],
                retrieve=row.retrieve.name if row.retrieve else None,
                aggregation=row.aggregation.name if row.aggregation else None,
            )
            base_mappings.append(mapping_apply)

        # restore original order:
        transformations = list(reversed(transformations))
        base_mappings = list(reversed(base_mappings))

        model.transformations.extend(transformations)
        model.mappings.extend(base_mappings)

        model.model_templates.append(
            ModelTemplateApply(
                external_id=f"ModelTemplate_{watercourse.name}",
                version=watercourse.config_version,
                shop_version=shop_version,
                watercourse=watercourse.name,
                model=file_ref.external_id,
                base_mappings=[m.external_id for m in base_mappings],
            )
        )

    return model


def _make_ext_id(watercourse_name: str, *args: str) -> str:
    hash_value = md5(watercourse_name.encode())
    for arg in args:
        hash_value.update(arg.encode())
    return f"Tr__{hash_value.hexdigest()}"


def _is_time_series(shop_attribute_value) -> bool:
    return isinstance(shop_attribute_value, dict) and isinstance(list(shop_attribute_value)[0], datetime)


def _to_model_without_timeseries(
    yaml_raw_path: str,
    non_time_series_attributes_to_remove: set[str] | None = None,
    encoding=None,
) -> dict:
    non_time_series_attributes_to_remove = non_time_series_attributes_to_remove or set()

    model_incl_time_and_connections = load_yaml(Path(yaml_raw_path), encoding=encoding, clean_data=True)

    model = model_incl_time_and_connections["model"]

    for object_type, objects in model.items():
        for object_attributes in objects.values():
            for attribute_name, attribute_data in list(
                object_attributes.items()
            ):  # Needs list() since we pop while iterating over the dict
                if f"{object_type}.{attribute_name}" in non_time_series_attributes_to_remove:
                    object_attributes.pop(attribute_name)

                elif _is_time_series(attribute_data):
                    object_attributes.pop(attribute_name)

    model_dict = {"model": model}

    # TODO: remove this when standardizing input SHOP file structure (model vs model + connections)
    if "connections" in model_incl_time_and_connections:
        model_dict["connections"] = model_incl_time_and_connections["connections"]

    return model_dict



from functools import cached_property

from cognite.client import CogniteClient


class DataSetsAPI:
    def __init__(self, client: CogniteClient, read: str, write: str):
        self.cdf = client
        self._read = read
        self._write = write

    @cached_property
    def read_dataset_id(self) -> int:
        return self._retrieve_with_raise_if_none(self._read).id

    @cached_property
    def write_dataset_id(self) -> int:
        return self._retrieve_with_raise_if_none(self._read).id

    def _retrieve_with_raise_if_none(self, dataset):
        dataset = self.cdf.data_sets.retrieve(dataset)
        if dataset is None:
            raise ValueError(f"Dataset {dataset} not found.")
        return dataset



import json
from dataclasses import dataclass
from typing import List, Optional, Union

from cognite.powerops.preprocessor.transformation_functions import Transformation, transformation_factory

from ..get_fdm_data import Mapping


@dataclass
class TimeSeriesMapping:
    shop_model_path: str
    transformations: List[Transformation]
    time_series_external_id: Optional[str] = None
    retrieve: Optional[str] = None
    aggregation: Optional[str] = None

    @classmethod
    def from_mapping_model(cls, mapping: Mapping) -> "TimeSeriesMapping":
        transformations = []
        if mapping.transformations is not None:
            for t in mapping.transformations.items:
                kwargs = None
                if isinstance(t.arguments, str):
                    kwargs = json.loads(t.arguments)
                kwargs = {} if kwargs is None else kwargs

                transformation = transformation_factory(t.method, kwargs=kwargs)
                transformations.append(transformation)

        return cls(
            shop_model_path=mapping.path,
            time_series_external_id=mapping.timeseries_external_id,
            transformations=transformations,
            retrieve=mapping.retrieve,
            aggregation=mapping.aggregation,
        )

    @classmethod
    def from_dict(cls, d: dict) -> "TimeSeriesMapping":
        transformations_string = "".join(
            d.get(key, "")
            for key in (
                "transformations",
                "transformations1",
                "transformations2",
                "transformations3",
            )
        )
        transformation_dicts = json.loads(transformations_string) if transformations_string else []
        transformations = [
            transformation_factory(transformation_type=d["transformation"], kwargs=d.get("kwargs", {}))
            for d in transformation_dicts
        ]
        return cls(
            shop_model_path=d["shop_model_path"],
            time_series_external_id=d.get("time_series_external_id"),
            transformations=transformations,
            retrieve=d.get("retrieve"),
            aggregation=d.get("aggregation"),
        )

    def split_shop_model_path(self) -> List[str]:
        return self.shop_model_path.split(".")

    @property
    def object_type(self) -> str:
        return self.split_shop_model_path()[0]

    @property
    def instance(self) -> Union[int, str]:
        instance = self.split_shop_model_path()[1]
        # TODO: confirm that this approach is OK
        if instance.isdigit():
            # E.g. market or scenario number
            return int(instance)
        return instance

    @property
    def attribute(self) -> str:
        return self.split_shop_model_path()[2]



from __future__ import annotations

from typing import Sequence, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.market_configuration.data_classes import Market, MarketApply, MarketList

from ._core import TypeAPI


class MarketsAPI(TypeAPI[Market, MarketApply, MarketList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "Market", "240c8b617aa6cd"),
            class_type=Market,
            class_apply_type=MarketApply,
            class_list=MarketList,
        )

    def apply(self, market: MarketApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = market.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(MarketApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(MarketApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Market:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> MarketList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Market | MarketList:
        if isinstance(external_id, str):
            return self._retrieve((self.sources.space, external_id))
        else:
            return self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> MarketList:
        return self._list(limit=limit)



from __future__ import annotations

from collections import defaultdict
from typing import Dict, List, Sequence, Tuple, overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._constants import INSTANCES_LIST_LIMIT_DEFAULT

from cognite.powerops.clients.core.data_classes import Plant, PlantApply, PlantList

from ._core import TypeAPI


class PlantGeneratorsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Plant.generators"},
        )
        if isinstance(external_id, str):
            is_plant = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_plant))

        else:
            is_plants = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_plants))

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Plant.generators"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class PlantReservoirsAPI:
    def __init__(self, client: CogniteClient):
        self._client = client

    def retrieve(self, external_id: str | Sequence[str]) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Plant.inletReservoirs"},
        )
        if isinstance(external_id, str):
            is_plant = f.Equals(
                ["edge", "startNode"],
                {"space": "power-ops", "externalId": external_id},
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_plant))

        else:
            is_plants = f.In(
                ["edge", "startNode"],
                [{"space": "power-ops", "externalId": ext_id} for ext_id in external_id],
            )
            return self._client.data_modeling.instances.list("edge", limit=-1, filter=f.And(is_edge_type, is_plants))

    def list(self, limit=INSTANCES_LIST_LIMIT_DEFAULT) -> dm.EdgeList:
        f = dm.filters
        is_edge_type = f.Equals(
            ["edge", "type"],
            {"space": "power-ops", "externalId": "Plant.inletReservoirs"},
        )
        return self._client.data_modeling.instances.list("edge", limit=limit, filter=is_edge_type)


class PlantsAPI(TypeAPI[Plant, PlantApply, PlantList]):
    def __init__(self, client: CogniteClient):
        super().__init__(
            client=client,
            sources=dm.ViewId("power-ops", "Plant", "a92c056aae6d26"),
            class_type=Plant,
            class_apply_type=PlantApply,
            class_list=PlantList,
        )
        self.generators = PlantGeneratorsAPI(client)
        self.inlet_reservoirs = PlantReservoirsAPI(client)

    def apply(self, plant: PlantApply, replace: bool = False) -> dm.InstancesApplyResult:
        instances = plant.to_instances_apply()
        return self._client.data_modeling.instances.apply(nodes=instances.nodes, edges=instances.edges, replace=replace)

    def delete(self, external_id: str | Sequence[str]) -> dm.InstancesDeleteResult:
        if isinstance(external_id, str):
            return self._client.data_modeling.instances.delete(nodes=(PlantApply.space, external_id))
        else:
            return self._client.data_modeling.instances.delete(
                nodes=[(PlantApply.space, id) for id in external_id],
            )

    @overload
    def retrieve(self, external_id: str) -> Plant:
        ...

    @overload
    def retrieve(self, external_id: Sequence[str]) -> PlantList:
        ...

    def retrieve(self, external_id: str | Sequence[str]) -> Plant | PlantList:
        if isinstance(external_id, str):
            plant = self._retrieve((self.sources.space, external_id))

            generator_edges = self.generators.retrieve(external_id)
            plant.generators = [edge.end_node.external_id for edge in generator_edges]
            inlet_reservoir_edges = self.inlet_reservoirs.retrieve(external_id)
            plant.inlet_reservoirs = [edge.end_node.external_id for edge in inlet_reservoir_edges]

            return plant
        else:
            plants = self._retrieve([(self.sources.space, ext_id) for ext_id in external_id])

            generator_edges = self.generators.retrieve(external_id)
            self._set_generators(plants, generator_edges)
            inlet_reservoir_edges = self.inlet_reservoirs.retrieve(external_id)
            self._set_inlet_reservoirs(plants, inlet_reservoir_edges)

            return plants

    def list(self, limit: int = INSTANCES_LIST_LIMIT_DEFAULT) -> PlantList:
        plants = self._list(limit=limit)

        generator_edges = self.generators.list(limit=-1)
        self._set_generators(plants, generator_edges)
        inlet_reservoir_edges = self.inlet_reservoirs.list(limit=-1)
        self._set_inlet_reservoirs(plants, inlet_reservoir_edges)

        return plants

    @staticmethod
    def _set_generators(plants: Sequence[Plant], generator_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in generator_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for plant in plants:
            node_id = plant.id_tuple()
            if node_id in edges_by_start_node:
                plant.generators = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]

    @staticmethod
    def _set_inlet_reservoirs(plants: Sequence[Plant], inlet_reservoir_edges: Sequence[dm.Edge]):
        edges_by_start_node: Dict[Tuple, List] = defaultdict(list)
        for edge in inlet_reservoir_edges:
            edges_by_start_node[edge.start_node.as_tuple()].append(edge)

        for plant in plants:
            node_id = plant.id_tuple()
            if node_id in edges_by_start_node:
                plant.inlet_reservoirs = [edge.end_node.external_id for edge in edges_by_start_node[node_id]]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import DomainModel, DomainModelApply, InstancesApply, TypeList

if TYPE_CHECKING:
    from ._rkom_bids import RKOMBidApply
    from ._shop_transformations import ShopTransformationApply

__all__ = ["RKOMProces", "RKOMProcesApply", "RKOMProcesList"]


class RKOMProces(DomainModel):
    space: ClassVar[str] = "power-ops"
    bid: Optional[str] = None
    incremental_mapping: list[str] = Field([], alias="incrementalMapping")
    name: Optional[str] = None
    process_events: list[str] = Field([], alias="processEvents")
    shop: Optional[str] = None
    timezone: Optional[str] = None


class RKOMProcesApply(DomainModelApply):
    space: ClassVar[str] = "power-ops"
    bid: Optional[Union[str, "RKOMBidApply"]] = Field(None, repr=False)
    incremental_mapping: list[str] = []
    name: Optional[str] = None
    process_events: list[str] = []
    shop: Optional[Union[str, "ShopTransformationApply"]] = Field(None, repr=False)
    timezone: Optional[str] = None

    def _to_instances_apply(self, cache: set[str]) -> InstancesApply:
        if self.external_id in cache:
            return InstancesApply([], [])

        sources = []
        source = dm.NodeOrEdgeData(
            source=dm.ContainerId("power-ops", "RKOMProcess"),
            properties={
                "bid": {
                    "space": "power-ops",
                    "externalId": self.bid if isinstance(self.bid, str) else self.bid.external_id,
                },
                "incrementalMapping": self.incremental_mapping,
                "name": self.name,
                "processEvents": self.process_events,
                "shop": {
                    "space": "power-ops",
                    "externalId": self.shop if isinstance(self.shop, str) else self.shop.external_id,
                },
                "timezone": self.timezone,
            },
        )
        sources.append(source)

        this_node = dm.NodeApply(
            space=self.space,
            external_id=self.external_id,
            existing_version=self.existing_version,
            sources=sources,
        )
        nodes = [this_node]
        edges = []

        if isinstance(self.bid, DomainModelApply):
            instances = self.bid._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        if isinstance(self.shop, DomainModelApply):
            instances = self.shop._to_instances_apply(cache)
            nodes.extend(instances.nodes)
            edges.extend(instances.edges)

        return InstancesApply(nodes, edges)


class RKOMProcesList(TypeList[RKOMProces]):
    _NODE = RKOMProces



from __future__ import annotations

from typing import ClassVar, Optional

from pydantic import Field

from cognite.powerops.cdf_labels import AssetLabel
from cognite.powerops.resync.models.cdf_resources import CDFSequence

from .base import Bid, Market, Process, ShopTransformation


class DayAheadBid(Bid):
    is_default_config_for_price_area: bool
    main_scenario: str
    price_area: str
    price_scenarios: dict[str, str]
    no_shop: bool
    bid_process_configuration_name: str
    bid_matrix_generator_config_external_id: str
    market_config_external_id: str


class DayAheadProcess(Process):
    type_: ClassVar[str] = "POWEROPS_bid_process_configuration"
    label: ClassVar[AssetLabel] = AssetLabel.BID_PROCESS_CONFIGURATION
    parent_description: ClassVar[str] = "Configurations used in bid matrix generation processes"
    shop: ShopTransformation
    bid: DayAheadBid
    bid_matrix_generator_config: Optional[CDFSequence] = None
    incremental_mapping: list[CDFSequence] = Field(default_factory=list)


class NordPoolMarket(Market):
    max_price: float
    min_price: float
    price_steps: int
    price_unit: str
    tick_size: float
    time_unit: str
    trade_lot: float



from cognite.client import CogniteClient

from cognite.powerops.clients.cogshop import CogShopClient
from cognite.powerops.clients.data_set_api import DataSetsAPI

from .api.shop_api import ShopModelsAPI
from .api.shop_result_files_api import ShopFilesAPI
from .api.shop_results_api import ShopRunResultsAPI
from .api.shop_run_api import ShopRunsAPI


class ShopClient:
    def __init__(self, client: CogniteClient, cogshop: CogShopClient, data_set_api: DataSetsAPI, cogshop_version: str):
        self.models = ShopModelsAPI(cogshop)
        self.files = ShopFilesAPI(client)
        self.results = ShopRunResultsAPI(client, self.files)
        self.runs = ShopRunsAPI(client, data_set_api, self.results, cogshop_version)
